"""Playback for live streams."""

import operator
import re
import tempfile
from dataclasses import asdict, dataclass
from datetime import datetime, timedelta
from pathlib import Path
from typing import Iterable, Self, TypedDict, TypeGuard, Unpack
from urllib.parse import parse_qs, urlparse

import requests
import structlog
from platformdirs import user_cache_path

from ytpb.cache import read_from_cache, write_to_cache
from ytpb.download import (
    compose_default_segment_filename,
    download_segment,
    SegmentOutputFilename,
)
from ytpb.errors import (
    BaseUrlExpiredError,
    CachedItemNotFoundError,
    MaxRetryError,
    SequenceLocatingError,
)
from ytpb.fetchers import InfoFetcher, YtpbInfoFetcher
from ytpb.info import LEFT_NOT_FETCHED, LeftNotFetched, YouTubeVideoInfo
from ytpb.locate import SegmentLocator
from ytpb.representations import extract_representations
from ytpb.segment import Segment
from ytpb.streams import SetOfStreams, Streams
from ytpb.types import (
    AbsolutePointInStream,
    AudioOrVideoStream,
    PointInStream,
    RelativePointInStream,
    RelativeSegmentSequence,
    SegmentSequence,
    Timestamp,
)
from ytpb.utils.other import resolve_relativity_in_interval
from ytpb.utils.url import (
    build_video_url_from_base_url,
    check_base_url_is_expired,
    extract_id_from_video_url,
    extract_parameter_from_url,
)

logger = structlog.get_logger(__name__)

SEGMENT_URL_PATTERN = r"https://.+\.googlevideo\.com/videoplayback/.+"


@dataclass
class RewindTreeNode:
    key: Timestamp
    value: SegmentSequence
    left: Self | None = None
    right: Self | None = None


@dataclass
class RewindTreeMap:
    """A binary search tree implementation to store key-value pairs.

    Keys represent timestamps of segments, while values are sequence numbers.
    """

    root: RewindTreeNode | None = None

    @staticmethod
    def _is_tree_node(node: RewindTreeNode | None) -> TypeGuard[RewindTreeNode]:
        return node is not None

    @staticmethod
    def _insert(
        node: RewindTreeNode | None, key: Timestamp, value: SegmentSequence
    ) -> RewindTreeNode:
        if not RewindTreeMap._is_tree_node(node):
            return RewindTreeNode(key, value, None, None)
        if key < node.key:
            left = RewindTreeMap._insert(node.left, key, value)
            return RewindTreeNode(node.key, node.value, left, node.right)
        elif key > node.key:
            right = RewindTreeMap._insert(node.right, key, value)
            return RewindTreeNode(node.key, node.value, node.left, right)
        else:
            return RewindTreeNode(node.key, value, node.left, node.right)

    def insert(self, key: Timestamp, value: SegmentSequence) -> None:
        """Inserts a pair of timestamp and sequence number into the tree."""
        self.root = RewindTreeMap._insert(self.root, key, value)

    @staticmethod
    def _closest(
        node: RewindTreeNode | None, target: Timestamp, closest: RewindTreeNode
    ) -> RewindTreeNode | None:
        if not RewindTreeMap._is_tree_node(node):
            return closest
        if abs(target - closest.key) > abs(target - node.key):
            closest = node
        if target < node.key:
            return RewindTreeMap._closest(node.left, target, closest)
        elif target > node.key:
            return RewindTreeMap._closest(node.right, target, closest)
        else:
            return closest

    def closest(self, target: Timestamp) -> RewindTreeNode | None:
        """Finds the node closest to the target timestamp."""
        if self.root is None:
            return None
        return RewindTreeMap._closest(self.root, target, self.root)


@dataclass(frozen=True)
class RewindMoment:
    """Represents a moment that has been rewound."""

    #: An actual rewound date.
    date: datetime
    #: A sequence number of a segment with a target date.
    sequence: SegmentSequence
    #: A difference (in seconds) between segment start ingestion date and target
    #: date.
    cut_at: float
    #: Whether a moment represents the end of an interval.
    is_end: bool = False
    #: Whether a moment falls in gap.
    falls_in_gap: bool = False


@dataclass(frozen=True)
class RewindInterval:
    """Represents an interval that has been rewound."""

    start: RewindMoment
    end: RewindMoment

    def __post_init__(self):
        if self.start.sequence > self.end.sequence:
            raise ValueError(
                "Start moment is ahead of the end one: "
                f"{self.start.sequence} > {self.end.sequence}"
            )

    @property
    def duration(self) -> timedelta:
        """An interval duration."""
        return self.end.date - self.start.date

    @property
    def sequences(self) -> Iterable[SegmentSequence]:
        """Segment sequence numbers that represent the interval."""
        return range(self.start.sequence, self.end.sequence + 1)


class PlaybackSession(requests.Session):
    """A session to use with :class:`Playback`.

    This session can be used to provide a retry mechanism for failed
    requests. Here is a list of handled HTTP status codes for segment URLs:

    - 403: Refresh segment base URL, and repeat a request
    - 404: Retry a request with no change
    """

    max_retries: int = 3

    def __init__(self, playback: "Playback" = None, **kwargs):
        super().__init__(**kwargs)

        self.playback = playback
        self.hooks["response"].append(self._handle_http_errors)

    def set_playback(self, playback):
        self.playback = playback

    def _handle_403_error(self, request: requests.Request) -> None:
        old_corresponding_stream = next(
            iter(
                self.playback.streams.filter(
                    lambda x: request.url.startswith(x.base_url)
                )
            )
        )
        old_base_url = old_corresponding_stream.base_url

        self.playback.fetch_and_set_essential()
        new_corresponding_stream = self.playback.streams.get_by_itag(
            old_corresponding_stream.itag
        )
        new_base_url = new_corresponding_stream.base_url

        request.url = request.url.replace(old_base_url, new_base_url)

    def _handle_http_errors(
        self, response: requests.Response, *args, **kwargs
    ) -> requests.Response:
        if response.ok:
            return response

        request = response.request
        retries_count = getattr(request, "retries_count", 0)

        if retries_count < self.max_retries:
            if re.match(SEGMENT_URL_PATTERN, request.url):
                logger.debug("Received %s for %s", response.status_code, request.url)
                logger.debug(
                    "Handle error, and make another try (%s of %s)",
                    retries_count + 1,
                    self.max_retries,
                )
                match response.status_code:
                    case 403:
                        self._handle_403_error(request)
                    case 404:
                        pass
                    case _:
                        logger.debug("Unhandleable error encountered, do nothing")
                        return response
                request.retries_count = retries_count + 1
                return self.send(request, verify=False)
        else:
            raise MaxRetryError(
                f"Maximum number of retries exceeded with URL: {request.url}",
                response,
            )


class _PlaybackOptions(TypedDict, total=False):
    user_agent: str


class Playback:
    """The playback for live streams."""

    def __init__(
        self,
        video_url: str,
        session: requests.Session | None = None,
        fetcher: InfoFetcher | None = None,
        write_to_cache: bool = False,
        **kwargs: Unpack[_PlaybackOptions],
    ) -> None:
        """Constructs a playback.

        To work with a playback, fetch the essential information afterwards::

            playback = Playback(video_url)
            playback.fetch_and_set_essential()

        Args:
            video_url: A video URL.
            session: An instance of :class:`requests.Session`.
            fetcher: A fetcher used to gather the video information and streams.
            write_to_cache: Whether to write to cache.

        Keyword Args:
            user_agent: The HTTP User-Agent string used for requests. Note that
              this value has priority over the value from ``session``
              headers.
        """
        self.video_url = video_url
        self.fetcher = fetcher or YtpbInfoFetcher(video_url)
        self._need_to_cache = write_to_cache

        self.session = session or PlaybackSession(self)
        if user_agent := kwargs.get("user_agent"):
            self.session.headers["User-Agent"] = user_agent

        self._info: YouTubeVideoInfo | LeftNotFetched | None = None
        self._streams: SetOfStreams | None = None
        self._temp_directory: Path | None = None
        self._cache_directory: Path | None = None

        self.rewind_history = RewindTreeMap()

    @classmethod
    def from_url(cls, video_url: str, **kwargs) -> "Playback":
        """Creates a playback for the given video URL.

        This also fetches the video information and streams.

        Args:
            video_url: A video URL.
            **kwargs: Optional arguments that :class:`Playback` takes.

        Returns:
            An instance of :class:`Playback`.
        """
        playback = cls(video_url, **kwargs)
        playback.fetch_and_set_essential()
        return playback

    @classmethod
    def from_manifest(
        cls,
        manifest_path: Path,
        fetch_video_info: bool = True,
        **kwargs,
    ) -> "Playback":
        with open(manifest_path, "r", encoding="utf-8") as f:
            list_of_streams = extract_representations(f.read())
            streams = Streams(list_of_streams)

        some_base_url = next(iter(streams)).base_url
        video_url = build_video_url_from_base_url(some_base_url)

        if not check_base_url_is_expired(some_base_url):
            playback = cls(video_url, **kwargs)
            playback._set_streams(streams, fetch_video_info=fetch_video_info)
        else:
            logger.debug("Found expired base url, ignore the manifest")
            raise BaseUrlExpiredError

        return playback

    @classmethod
    def from_cache(cls, video_url: str, **kwargs) -> "Playback":
        """Creates a playback for the given URL from cache.

        This also implies writing to cache. Note that an unexpired cache item
        for a stream should exist.

        Example:
            To write a new cache item, create a playback with
            ``write_to_cache=True``::

                from ytpb.errors import CachedItemNotFoundError
                try:
                    playback = Playback.from_cache(video_url)
                except CachedItemNotFoundError:
                    playback = Playback.from_url(video_url, write_to_cache=True)

        Args:
            video_url: A video URL.

        Raises:
            CachedItemNotFoundError: If a cache item is not found or expired.
        """
        video_id = extract_id_from_video_url(video_url)
        cached_item = read_from_cache(video_id, Playback.get_cache_directory())
        if cached_item is None:
            raise CachedItemNotFoundError

        playback = cls(video_url, write_to_cache=True, **kwargs)
        playback._info = YouTubeVideoInfo(**cached_item["info"])
        playback._streams = Streams.from_dicts(cached_item["streams"])

        return playback

    @staticmethod
    def get_cache_directory():
        """Gets the cache directory."""
        return user_cache_path("ytpb")

    def get_temp_directory(self) -> Path:
        """Gets the run temporary directory."""
        if self._temp_directory is None:
            self._temp_directory = Path(tempfile.mkdtemp(prefix="ytpb-"))
            logger.debug("Run temp directory set to %s", self._temp_directory)
        return self._temp_directory

    @property
    def video_url(self) -> str:
        return self._video_url

    @video_url.setter
    def video_url(self, value: str) -> None:
        self._video_url = value
        try:
            self._video_id = parse_qs(urlparse(value).query)["v"][0]
        except (KeyError, IndexError):
            print("Could not extract video ID from URL")

    @property
    def video_id(self) -> str:
        return self._video_id

    @property
    def info(self) -> YouTubeVideoInfo | LeftNotFetched:
        if self._info is None:
            raise ValueError(
                "Attribute 'info' is not set, call 'fetch_and_set_essential' first"
            )
        return self._info

    @property
    def streams(self) -> SetOfStreams:
        if self._streams is None:
            raise ValueError(
                "Attribute 'streams' is not set, call 'fetch_and_set_essential' first"
            )
        return self._streams

    @property
    def locations(self) -> dict[str, Path]:
        temp_directory = self.get_temp_directory()
        return {
            ".": temp_directory,
            "segments": temp_directory / "segments",
        }

    def _set_streams(self, value: SetOfStreams, fetch_video_info: bool = True) -> None:
        """Sets streams manually.

        By default, it also fetches information about a video.

        Args:
            value: Streams to set.
            fetch_video_info: Whether fetch information about a video or not. If
              not, the :attr:`.info` attribute will be set to
              :attr:`ytpb.info.LEFT_NOT_FETCHED`.

        Notes:
            In most cases, you don't need this. It could be only useful, when
            you can't fetch streams the usual way, with
            :meth:`fetch_and_set_essential`.
        """
        self._streams = value
        if not self._info or fetch_video_info:
            self._fetch_and_set_video_info()
        else:
            self._info = LEFT_NOT_FETCHED
        self._write_to_cache_if_needed()

    def _write_to_cache_if_needed(self):
        if self._need_to_cache:
            item_to_cache = {
                "info": asdict(self._info),
                "streams": [asdict(stream) for stream in self.streams],
            }
            some_base_url = next(iter(self.streams)).base_url
            expires_at = extract_parameter_from_url("expire", some_base_url)
            cache_directory = Playback.get_cache_directory()
            write_to_cache(self.video_id, expires_at, item_to_cache, cache_directory)

    def _fetch_and_set_video_info(self) -> None:
        self._info = self.fetcher.fetch_video_info()

    def _fetch_and_set_streams(self) -> None:
        self._streams = self.fetcher.fetch_streams()

    def fetch_and_set_essential(self) -> None:
        """Fetches and sets essential information.

        Such information includes information about a video and streams.
        """
        self._fetch_and_set_video_info()
        self._fetch_and_set_streams()
        self._write_to_cache_if_needed()

    def download_segment(
        self,
        sequence: SegmentSequence,
        stream: AudioOrVideoStream,
        output_directory: Path | None = None,
        output_filename: (
            SegmentOutputFilename | None
        ) = compose_default_segment_filename,
        force_download: bool = False,
    ) -> Path:
        """Downloads a segment.

        Examples:
            Download a segment to a file and read it::

                from ytpb.segment import Segment
                downloaded_path = playback.download_segment(
                    0, playback.streams.get_by_itag("140")
                )
                segment = Segment.from_file(downloaded_path)

        Args:
            sequence: A segment sequence number.
            stream: A stream to which segment belongs.
            output_directory: Where to download a segment.
            output_filename: A segment output filename.
            force_download: Whether to force download a segment even if it exists.

        Returns:
            A path to the downloaded segment.
        """
        if output_directory is None:
            output_directory = self.locations["."]
        path = download_segment(
            sequence,
            stream.base_url,
            output_directory=output_directory,
            output_filename=output_filename,
            session=self.session,
            force_download=force_download,
        )
        return path

    def get_segment(
        self,
        sequence: SegmentSequence,
        stream: AudioOrVideoStream,
        segment_directory: Path | None = None,
        segment_filename: (
            SegmentOutputFilename | None
        ) = compose_default_segment_filename,
        download: bool = True,
    ) -> Segment:
        """Gets a :class:`Segment` representing a downloaded segment.

        By default, if a segment file cannot be found, it will be downloaded.

        Notes:
            See also :meth:`ytpb.segment.Segment.from_file`.

        Args:
            sequence: A segment sequence number.
            stream: A stream to which segment belongs.
            segment_directory: Where a segment is stored.
            segment_filename: A segment filename.
            download: Whether to download a segment if it doesn't exist.

        Returns:
            A :class:`Segment` object.

        Raises:
            FileNotFoundError: If couldn't find a downloaded segment, when
              download is not requested.
        """
        segment_directory = segment_directory or self.get_temp_directory()
        if callable(segment_filename):
            segment_filename_value = segment_filename(sequence, stream.base_url)
            segment_path = segment_directory / segment_filename_value
        else:
            segment_path = segment_directory / segment_filename

        try:
            segment = Segment.from_file(segment_path)
        except FileNotFoundError as exc:
            if download:
                downloaded_path = self.download_segment(sequence, stream)
                segment = Segment.from_file(downloaded_path)
            else:
                exc.add_note(
                    "Couldn't find a segment. Make sure to download it before "
                    "and the same segment filename is used"
                )
                raise
        return segment

    def locate_moment(
        self,
        point: AbsolutePointInStream,
        stream: AudioOrVideoStream | None = None,
        is_end: bool = False,
    ) -> RewindMoment:
        """Locates a moment by a point in stream.

        Args:
            point: An absolute point.
            stream: A stream to which segments used in locating belong. If not
              set, the first available stream will be used.
            is_end: Whether a moment represents the end of an interval.

        Notes:
            See also :class:`ytpb.locate.SegmentLocator`.

        Returns:
            A located moment of :class:`RewindMoment`.
        """
        stream = stream or next(iter(self.streams))

        def _get_non_located_date(segment: Segment) -> datetime:
            if is_end:
                date = segment.ingestion_end_date
            else:
                date = segment.ingestion_start_date
            return date

        segment: Segment | None = None

        match point:
            case SegmentSequence() as sequence:
                segment = Segment.from_file(self.download_segment(sequence, stream))
                date = _get_non_located_date(segment)
                moment = RewindMoment(date, sequence, 0, is_end)
            case datetime() as date:
                reference_sequence: SegmentSequence | None = None
                if reference := self.rewind_history.closest(date.timestamp()):
                    reference_sequence = reference.value
                sl = SegmentLocator(
                    stream.base_url,
                    reference_sequence=reference_sequence,
                    temp_directory=self.get_temp_directory(),
                    session=self.session,
                )
                locate_result = sl.find_sequence_by_time(date.timestamp(), end=is_end)

                if locate_result.falls_in_gap:
                    segment = self.get_segment(locate_result.sequence, stream)
                    date = _get_non_located_date(segment)
                    cut_at = 0
                else:
                    cut_at = locate_result.time_difference

                moment = RewindMoment(
                    date=date,
                    sequence=locate_result.sequence,
                    cut_at=cut_at,
                    is_end=is_end,
                    falls_in_gap=locate_result.falls_in_gap,
                )

        if segment is None:
            segment = self.get_segment(moment.sequence, stream)
        self.rewind_history.insert(
            segment.ingestion_start_date.timestamp(), moment.sequence
        )

        return moment

    def locate_interval(
        self,
        start_point: PointInStream,
        end_point: PointInStream,
        stream: AudioOrVideoStream | None = None,
    ) -> RewindInterval:
        """Locates an interval by start and end points in stream.

        Args:
            start_point: A start absolute or relative point.
            end_point: An end absolute or relative point.
            stream: A stream to which segments used in locating belong. If not
              set, the first available stream will be used.

        Notes:
            See also :class:`ytpb.locate.SegmentLocator`.

        Returns:
           A located interval of :class:`RewindInterval`.
        """
        # First, resolve relativity where it's possible, without locating:
        start_point, end_point = resolve_relativity_in_interval(start_point, end_point)

        stream = stream or next(iter(self.streams))
        if type(start_point) is type(end_point) is SegmentSequence:
            start_moment = self.locate_moment(start_point, stream=stream)
            end_moment = self.locate_moment(end_point, stream=stream, is_end=True)
            return RewindInterval(start_moment, end_moment)

        # Then, locate end if start is relative:
        if isinstance(start_point, RelativePointInStream):
            end_moment = self.locate_moment(end_point, stream=stream, is_end=True)
        else:
            end_moment = None

        # Finally, iterate over start and end points to locate moments:
        start_and_end_moments: list[RewindMoment | None] = [None, end_moment]
        for index, (point, is_start) in enumerate(
            ((start_point, True), (end_point, False))
        ):
            start_moment, end_moment = start_and_end_moments
            if isinstance(point, RelativePointInStream):
                start_moment, end_moment = start_and_end_moments
                # Given the current relative point as delta, the resulted point:
                # start/end point = end/start point -/+ delta.
                if is_start:
                    contrary_moment = end_moment
                    contrary_op = operator.sub
                else:
                    contrary_moment = start_moment
                    contrary_op = operator.add

                match point:
                    case RelativeSegmentSequence() as sequence_delta:
                        sequence = contrary_op(contrary_moment.sequence, sequence_delta)
                        moment = self.locate_moment(sequence, stream, not is_start)
                    case timedelta() as time_delta:
                        date = contrary_op(contrary_moment.date, time_delta)
                        moment = self.locate_moment(date, stream, not is_start)
            else:
                if end_moment:
                    continue
                moment = self.locate_moment(point, stream, not is_start)

            start_and_end_moments[index] = moment

        try:
            resulted_interval = RewindInterval(*start_and_end_moments)
        except ValueError as exc:
            raise SequenceLocatingError(str(exc)) from exc

        return resulted_interval



import logging
import operator
import os
import tomllib
from collections import ChainMap
from collections.abc import Hashable, Mapping, MutableMapping
from copy import deepcopy
from functools import reduce
from pathlib import Path
from typing import Any

import platformdirs
import structlog

from ytpb.cli.formats import ALIASES as FORMAT_ALIASES


class AddressableMixin:
    def traverse(self, address: str, default: Any = None, delimiter: str = ".") -> Any:
        """Traverse nested dictionary to access a value by an address."""
        try:
            keys = address.split(delimiter)
            return reduce(operator.getitem, keys, self)
        except KeyError:
            return default


class AddressableDict(dict, AddressableMixin):
    """A dictionary that allows to use an address to access a nested value."""


class DeepChainMap(ChainMap):
    """A ChainMap that works on nested dictionaries."""

    def __getitem__(self, key: Hashable) -> Any:
        if not isinstance(value := super().__getitem__(key), Mapping):
            return value

        values: list[dict] = []
        for mapping in self.maps:
            try:
                values.append(mapping[key])
            except KeyError:
                pass

        if values:
            return self.__class__(*values)

        return self.__missing__(key)


class AddressableChainMap(DeepChainMap, AddressableMixin):
    """A ChainMap that allows to use an address to access a nested value."""


USER_AGENT = "Mozilla/5.0 (Android 14; Mobile; rv:68.0) Gecko/68.0 Firefox/120.0"


DEFAULT_OUTPUT_PATH = "{{ title|adjust }}_{{ id }}_{{ input_start_date|isodate }}"

DEFAULT_CONFIG = AddressableDict(
    {
        "version": 2,
        "options": {
            "download": {
                "audio_format": "itag eq 140",
                "video_format": (
                    "("
                    "format eq mp4 and [height eq 720 or height eq 1080] and frame_rate eq 30 "
                    "?: format eq mp4 and height le 1080"
                    "?: height le 1080"
                    ")"
                    " | best"
                ),
                "output_path": DEFAULT_OUTPUT_PATH,
            },
            "capture": {
                "frame": {
                    "video_format": "(height ge 1080 and frame_rate eq 30 ?: all) | best",
                    "output_path": "{{ title|adjust }}_{{ id }}_{{ moment_date|isodate }}.jpg",
                },
                "timelapse": {
                    "video_format": "(height ge 1080 and frame_rate eq 30 ?: all) | best",
                    "output_path": (
                        "{{ title|adjust }}_{{ id }}/{{ input_start_date|isodate }}/{{ (every | duration).replace('PT', 'ET') }}/"
                        "{{ title|adjust }}_{{ id }}_{{ input_start_date|isodate }}_{{ (every | duration).replace('PT', 'ET') }}_%04d.jpg"
                    ),
                },
            },
            "mpd": {
                "compose": {
                    "audio_formats": "itag eq 140",
                    "video_formats": (
                        "codecs eq vp9 and [height eq 1080 or height eq 720] and frame_rate eq 30 ?:"
                        "codecs eq vp9 and [height eq 1080 or height eq 720]"
                    ),
                    "output_path": f"{DEFAULT_OUTPUT_PATH}.mpd",
                }
            },
        },
        "general": {
            "preview_duration": 10,
            "user_agent": USER_AGENT,
            "aliases": FORMAT_ALIASES,
        },
        "output": {
            "metadata": {
                "dates": "iso",
            },
        },
    }
)


def setup_logging(level: int) -> None:
    structlog.configure(
        processors=[
            structlog.processors.add_log_level,
            structlog.processors.StackInfoRenderer(),
            structlog.processors.TimeStamper(fmt="iso", utc=True),
            structlog.dev.ConsoleRenderer(
                exception_formatter=structlog.dev.plain_traceback,
                # For details about NO_COLOR, see https://no-color.org/
                colors="NO_COLOR" not in os.environ,
            ),
            structlog.dev.set_exc_info,
        ],
        wrapper_class=structlog.make_filtering_bound_logger(level),
        context_class=dict,
        logger_factory=structlog.PrintLoggerFactory(),
        cache_logger_on_first_use=True,
    )

    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("urllib3").propagate = False


def get_default_config_path() -> Path:
    return platformdirs.user_config_path() / "ytpb/config.toml"


def load_config_from_file(path: Path) -> dict:
    with path.open("rb") as f:
        return tomllib.load(f)


def update_nested_dict(base: MutableMapping, updates: MutableMapping) -> MutableMapping:
    """Update a base nested dict with values from updates. The new, update
    dictionary will be returned.
    """
    updated_dict = deepcopy(base)
    for key, value in updates.items():
        if isinstance(value, Mapping) and isinstance(base.get(key), Mapping):
            updated_dict[key] = update_nested_dict(base[key], value)
        else:
            updated_dict[key] = value
    return updated_dict



import shlex
import subprocess
from pathlib import Path
from typing import Any

import structlog

from ytpb.errors import FFmpegRunError

logger = structlog.get_logger(__name__)


def run_ffmpeg(
    args: str | list[str | Path], **subprocess_kwargs: Any
) -> subprocess.CompletedProcess:
    command = ["ffmpeg", "-v", "error", "-hide_banner", "-y"]
    if isinstance(args, str):
        command.extend(shlex.split(args))
    else:
        command.extend(args)

    logger.debug(" ".join([str(x) for x in command]))
    cp = subprocess.run(command, **subprocess_kwargs)
    try:
        cp.check_returncode()
    except subprocess.CalledProcessError as e:
        logger.error("There was a problem running FFmpeg command.")
        raise FFmpegRunError from e

    return cp



import datetime

from ytpb._version import __version__


project = "Ytpb"
author = "Maxim Stolyarchuk"
copyright = f"{datetime.date.today().year}, Maxim Stolyarchuk"
version = __version__
release = version

templates_path = ["_templates"]

html_sidebars = {
    "**": [
        "sidebarintro.html",
        "navigation.html",
        "searchbox.html",
    ]
}

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx_toolbox.collapse",
    "myst_parser",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
}

autosectionlabel_prefix_document = True
suppress_warnings = ["autosectionlabel.*"]

autodoc_default_options = {
    "members": True,
    "undoc-members": True,
    "show-inheritance": True,
    "autoclass_content": "both",
}

autodoc_preserve_defaults = True
autodoc_inherit_docstrings = True
autodoc_member_order = "bysource"

napoleon_google_docstring = True
napoleon_include_init_with_doc = True
napoleon_include_special_with_doc = True
napoleon_use_rtype = False
napoleon_use_ivars = False

html_theme = "alabaster"
html_theme_options = {
    "fixed_sidebar": True,
}



import pickle
import shutil
import sys
from datetime import datetime, timedelta, timezone
from functools import partial
from pathlib import Path
from typing import Any, Callable, TypedDict

import click
import cloup
import structlog
from cloup.constraints import constraint, mutually_exclusive, require_any

from ytpb import actions
from ytpb.cli.common import (
    create_playback,
    echo_notice,
    find_earliest_sequence,
    get_parameter_by_name,
    print_summary_info,
    query_streams_or_exit,
    raise_for_sequence_ahead_of_current,
    raise_for_too_far_sequence,
    resolve_output_path,
    stream_argument,
    suppress_output,
)
from ytpb.cli.options import (
    cache_options,
    interval_option,
    keep_temp_option,
    validate_output_path,
    yt_dlp_option,
)
from ytpb.cli.parameters import FormatSpecParamType, FormatSpecType, InputRewindInterval
from ytpb.cli.templating import (
    AudioStreamOutputPathContext,
    check_is_template,
    IntervalOutputPathContext,
    MinimalOutputPathContext,
    render_template,
    VideoStreamOutputPathContext,
)
from ytpb.cli.utils.path import (
    remove_directories_between,
    sanitize_for_filename,
    try_get_relative_path,
)
from ytpb.download import compose_default_segment_filename
from ytpb.errors import SequenceLocatingError
from ytpb.merge import merge_segments
from ytpb.types import (
    AudioOrVideoStream,
    DateInterval,
    RelativeSegmentSequence,
    SegmentSequence,
)
from ytpb.utils.other import resolve_relativity_in_interval
from ytpb.utils.remote import request_reference_sequence
from ytpb.utils.url import build_segment_url, extract_parameter_from_url

logger = structlog.get_logger(__name__)


class DownloadOutputPathContext(
    MinimalOutputPathContext,
    AudioStreamOutputPathContext,
    VideoStreamOutputPathContext,
    IntervalOutputPathContext,
): ...


class MetadataTagsContext(TypedDict):
    #: Video's title.
    title: str
    #: Video's author (channel's name).
    author: str
    #: YouTube video URL as a comment.
    comment: str
    #: Input start date.
    input_start_date: str
    #: Input end date.
    input_end_date: str
    #: Actual start date.
    actual_start_date: str
    #: Actual end date.
    actual_end_date: str
    #: Start segment sequence number.
    start_segment: str
    #: End segment sequence number.
    end_segment: str


def compose_resume_filename(
    video_id: str, streams: list[AudioOrVideoStream | None]
) -> str:
    for i, arg in enumerate(sys.argv):
        if arg in ("--interval", "-i"):
            interval_option_value = sys.argv[i + 1]
            break

    interval_part = interval_option_value
    for char in "-:@":
        interval_part = interval_part.replace(char, "")
    interval_part = interval_part.replace("/", "-")

    itag_part = "".join([stream.itag for stream in streams if stream])

    return f"{video_id}-{interval_part}-{itag_part}.resume"


@cloup.command("download", short_help="Download excerpts.", help="Download an excerpt.")
@cloup.option_group(
    "Input options",
    interval_option,
    cloup.option(
        "-af",
        "--audio-format",
        metavar="SPEC",
        type=FormatSpecParamType(FormatSpecType.AUDIO),
        help="Audio format to download.",
    ),
    cloup.option(
        "-vf",
        "--video-format",
        metavar="SPEC",
        type=FormatSpecParamType(FormatSpecType.VIDEO),
        help="Video format to download.",
    ),
    cloup.option(
        "-ps",
        "--preview-start",
        help="Preview interval start.",
        is_flag=True,
    ),
    cloup.option(
        "-pe",
        "--preview-end",
        help="Preview interval end.",
        is_flag=True,
    ),
)
@cloup.option_group(
    "Dump options",
    cloup.option(
        "--dump-base-urls",
        is_flag=True,
        help="Print base URLs and exit.",
    ),
    cloup.option(
        "--dump-segment-urls",
        is_flag=True,
        help="Print segment URLs and exit.",
    ),
    cloup.option(
        "--dump-rewind-interval",
        is_flag=True,
        help="Print segments rewind interval and exit.",
    ),
)
@cloup.option_group(
    "Output options",
    cloup.option(
        "-o",
        "--output",
        "output_path",
        type=click.Path(path_type=Path),
        help="Output path (without extension).",
        callback=validate_output_path(DownloadOutputPathContext),
    ),
    cloup.option(
        "-S", "--keep-segments", is_flag=True, help="Keep downloaded segments."
    ),
    cloup.option(
        "-s",
        "--segments-output-dir",
        "segments_output_dir_option",
        type=click.Path(path_type=Path),
        help="Location where to download segments to.",
    ),
    cloup.option(
        "-c", "--cut", is_flag=True, help="Accurately cut an excerpt at boundaries."
    ),
    cloup.option("--no-metadata", is_flag=True, help="Do not write metadata tags."),
    cloup.option(
        "--no-merge",
        is_flag=True,
        help="Only download segments, without merging.",
    ),
)
@click.option(
    "-x",
    "--dry-run",
    count=True,
    help="Run until locating (-xx) or downloading (-x).",
)
@yt_dlp_option
@click.option(
    "--ignore-resume", is_flag=True, help="Avoid resuming unfinished download."
)
@cache_options
@keep_temp_option
@stream_argument
@constraint(require_any, ["audio_format", "video_format"])
@constraint(mutually_exclusive, ["preview_start", "preview_end"])
@click.pass_context
def download_command(
    ctx: click.Context,
    interval: InputRewindInterval,
    audio_format: str,
    video_format: str,
    preview_start: bool,
    preview_end: bool,
    dump_base_urls: bool,
    dump_segment_urls: bool,
    dump_rewind_interval: bool,
    output_path: Path,
    keep_segments: bool,
    segments_output_dir_option: Path,
    cut: bool,
    no_metadata: bool,
    no_merge: bool,
    dry_run: bool,
    yt_dlp: bool,
    ignore_resume: bool,
    no_cache: bool,
    force_update_cache: bool,
    keep_temp: bool,
    stream_url: str,
) -> int:
    if dump_base_urls or dump_segment_urls or dump_rewind_interval:
        suppress_output()

    if no_merge:
        keep_segments = True

    playback = create_playback(ctx)

    queried_audio_streams = []
    if audio_format:
        logger.debug("Query audio stream by format spec", spec=audio_format)
        streams_to_query = playback.streams.filter(lambda x: x.type == "audio")
        queried_audio_streams = query_streams_or_exit(
            streams_to_query, audio_format, "--audio-format", allow_many=False
        )

    queried_video_streams = []
    if video_format:
        logger.debug("Query video stream by format spec", spec=video_format)
        streams_to_query = playback.streams.filter(lambda x: x.type == "video")
        queried_video_streams = query_streams_or_exit(
            streams_to_query, video_format, "--video-format", allow_many=False
        )

    if queried_audio_streams and queried_video_streams:
        click.echo("These representations will be downloaded:")
    else:
        click.echo("This representation will be downloaded:")

    audio_stream, video_stream = None, None
    if audio_format:
        (audio_stream,) = queried_audio_streams
        click.echo(
            f"   - Audio: itag {audio_stream.itag}, "
            f"{audio_stream.format} ({audio_stream.codecs}), "
            f"{audio_stream.audio_sampling_rate} Hz"
        )
        logger.info("Queried audio stream", base_url=audio_stream.base_url)

    if video_format:
        (video_stream,) = queried_video_streams
        click.echo(
            f"   - Video: itag {video_stream.itag}, "
            f"{video_stream.format} ({video_stream.codecs}), "
            f"{video_stream.width}x{video_stream.height}, "
            f"{video_stream.frame_rate} fps"
        )
        logger.info("Queried video stream", base_url=video_stream.base_url)

    if dump_base_urls:
        if audio_format:
            click.echo(audio_stream.base_url, ctx.obj.original_stdout)
        if video_format:
            click.echo(video_stream.base_url, ctx.obj.original_stdout)
        sys.exit()

    if dry_run == 2:
        echo_notice("This is a dry run, just exit.")
        sys.exit()

    click.echo()

    click.echo("(<<) Locating start and end in the stream... ", nl=False)

    reference_stream = video_stream or audio_stream
    reference_base_url = reference_stream.base_url
    head_sequence = request_reference_sequence(reference_base_url, playback.session)

    requested_start, requested_end = resolve_relativity_in_interval(*interval)

    if requested_start == "earliest":
        head_date = datetime.now(timezone.utc)
        requested_start = find_earliest_sequence(playback, head_sequence, head_date)

    if isinstance(requested_start, SegmentSequence):
        raise_for_too_far_sequence(
            requested_start, head_sequence, reference_base_url, ctx, "interval"
        )
        raise_for_sequence_ahead_of_current(
            requested_start, head_sequence, ctx, "interval"
        )

    match requested_end:
        case SegmentSequence() as x:
            raise_for_sequence_ahead_of_current(x, head_sequence, ctx, "interval")
        case "now":
            requested_end = head_sequence - 1

    preview_mode = preview_start or preview_end
    if not preview_mode and (requested_start == ".." or requested_end == ".."):
        raise click.BadParameter(
            "Open interval is only valid in the preview mode.",
            param=get_parameter_by_name("interval", ctx),
        )

    if preview_mode:
        preview_duration_value = ctx.obj.config.traverse("general.preview_duration")
        segment_duration = float(extract_parameter_from_url("dur", reference_base_url))
        preview_segments = round(preview_duration_value / segment_duration)

        if preview_start:
            if requested_start == "..":
                raise click.BadParameter(
                    "Start '..' is only valid in the end preview mode.",
                    param=get_parameter_by_name("interval", ctx),
                )
            elif isinstance(requested_start, timedelta):
                located_moment = playback.locate_moment(
                    requested_end, reference_stream, is_end=True
                )
                requested_start = located_moment.date - requested_start
            requested_end = RelativeSegmentSequence(preview_segments)
        if preview_end:
            if requested_end == "..":
                raise click.BadParameter(
                    "End '..' is only valid in the start preview mode.",
                    param=get_parameter_by_name("interval", ctx),
                )
            elif isinstance(requested_end, timedelta):
                located_moment = playback.locate_moment(
                    requested_start, reference_stream
                )
                requested_end = located_moment.date + requested_end
            requested_start = RelativeSegmentSequence(preview_segments)

    resume_run: bool = False
    resume_file_path = Path.cwd() / compose_resume_filename(
        playback.video_id, (audio_stream, video_stream)
    )

    if not ignore_resume and resume_file_path.exists():
        logger.debug("Load resume file from %s", resume_file_path)
        with open(resume_file_path, "rb") as f:
            resume_run = True
            pickled = pickle.load(f)
            rewind_interval = pickled["interval"]
            previous_segments_output_directory = Path.absolute(
                pickled["segments_output_directory"]
            )

    if not resume_run:
        try:
            # assert False, requested_start
            rewind_interval = playback.locate_interval(
                requested_start,
                requested_end,
                reference_stream,
            )
        except SequenceLocatingError:
            message = "\nerror: An error occured during segment locating, exit."
            click.echo(message, err=True)
            sys.exit(1)

    click.echo("done.")

    if dump_segment_urls:
        range_part = "[{start}-{end}]".format(
            start=rewind_interval.start.sequence, end=rewind_interval.end.sequence
        )
        build_dump_url = lambda base_url: build_segment_url(base_url, range_part)
        if audio_format:
            click.echo(build_dump_url(audio_stream.base_url), ctx.obj.original_stdout)
        if video_format:
            click.echo(build_dump_url(video_stream.base_url), ctx.obj.original_stdout)
        sys.exit()

    if dump_rewind_interval:
        click.echo(
            "{start}-{end}".format(
                start=rewind_interval.start.sequence, end=rewind_interval.end.sequence
            ),
            ctx.obj.original_stdout,
        )
        sys.exit()

    # if preview_mode and interval[1] != "..":
    if preview_mode:
        echo_notice(
            "Preview mode enabled, interval {} is ignored.".format(
                "end" if preview_start else "start"
            )
        )

    start_segment = playback.get_segment(
        rewind_interval.start.sequence, reference_stream
    )
    end_segment = playback.get_segment(rewind_interval.end.sequence, reference_stream)

    requested_start_date: datetime
    match requested_start:
        case SegmentSequence():
            requested_start_date = start_segment.ingestion_start_date
        case datetime():
            requested_start_date = requested_start
        case timedelta() as delta:
            requested_start_date = end_segment.ingestion_end_date - delta

    requested_end_date: datetime
    match requested_end:
        case SegmentSequence():
            requested_end_date = end_segment.ingestion_end_date
        case datetime():
            requested_end_date = requested_end
        case timedelta() as delta:
            requested_end_date = start_segment.ingestion_start_date + delta

    requested_date_interval = DateInterval(requested_start_date, requested_end_date)
    actual_date_interval = DateInterval(
        start_segment.ingestion_start_date,
        end_segment.ingestion_end_date,
    )

    cut_kwargs: dict[str, float] = {}
    if cut:
        cut_at_start = 0 if preview_start else rewind_interval.start.cut_at
        cut_at_end = 0 if preview_end else rewind_interval.end.cut_at
        cut_kwargs.update(
            {
                "cut_at_start": cut_at_start,
                "cut_at_end": cut_at_end,
            }
        )
        actual_date_interval = DateInterval(
            rewind_interval.start.date, rewind_interval.end.date
        )

    print_summary_info(requested_date_interval, actual_date_interval, rewind_interval)
    click.echo()

    if dry_run:
        echo_notice("This is a dry run. Skip downloading and exit.")
    else:
        # Absolute output path of an excerpt without extension.
        final_output_path: Path

        if preview_mode:
            default_output_path = ctx.obj.config.traverse(
                "options.download.output_path"
            )
            if str(output_path) == default_output_path:
                output_path = Path("{{ title|adjust }}_{{ id }}_preview")

        if check_is_template(str(output_path)):
            input_timezone = requested_date_interval.start.tzinfo
            template_context: DownloadOutputPathContext = {
                "id": playback.video_id,
                "title": sanitize_for_filename(playback.info.title),
                "audio_stream": audio_stream,
                "video_stream": video_stream,
                "input_start_date": requested_date_interval.start,
                "input_end_date": requested_date_interval.end,
                "actual_start_date": (
                    actual_date_interval.start.astimezone(input_timezone)
                ),
                "actual_end_date": actual_date_interval.end.astimezone(input_timezone),
                "duration": requested_end_date - requested_start_date,
            }
            final_output_path = render_template(
                output_path,
                ctx.obj.jinja_environment,
                template_context,
            )
        else:
            final_output_path = output_path
        final_output_path = resolve_output_path(final_output_path)

        if preview_mode:
            segments_output_directory = playback.get_temp_directory()
        elif segments_output_dir_option:
            segments_output_directory = segments_output_dir_option
        elif keep_segments or no_merge:
            segments_output_directory = final_output_path
        else:
            segments_output_directory = resume_file_path.with_name(
                f"{resume_file_path.stem}"
            )
        segments_output_directory = segments_output_directory.absolute()
        logger.debug("Segments output directory set to %s", segments_output_directory)

        if resume_run:
            click.echo(
                f"~ Found unfinished download, continue from {resume_file_path.name}"
            )
            click.echo()
            if segments_output_directory != previous_segments_output_directory:
                click.echo(
                    "fatal: The previous segments output directory is not "
                    "the same as the current run:\n'{}' != '{}'.".format(
                        try_get_relative_path(previous_segments_output_directory),
                        try_get_relative_path(segments_output_directory),
                    ),
                    err=True,
                )
                click.echo(
                    "\nUse '--segments-output-dir {}' or '--ignore-resume'.".format(
                        try_get_relative_path(previous_segments_output_directory)
                    ),
                    err=True,
                )
                sys.exit(1)

        if segments_output_dir_option is None or not segments_output_directory.exists():
            need_to_remove_segments_directory = True
            segments_directory_to_remove = segments_output_directory
            for directory in segments_output_directory.parents:
                if directory.exists():
                    break
                segments_directory_to_remove = directory
            segments_output_directory.mkdir(parents=True, exist_ok=True)
        else:
            need_to_remove_segments_directory = False

        if not resume_run and not preview_mode:
            with open(resume_file_path, "wb") as f:
                logger.debug("Write resume file to %s", resume_file_path)
                to_pickle = {
                    "interval": rewind_interval,
                    "segments_output_directory": segments_output_directory,
                }
                pickle.dump(to_pickle, f)

        if resume_run:
            extract_sequence_number = lambda p: int(p.name.split(".")[0])
            latest_sequence_number = extract_sequence_number(
                sorted(
                    segments_output_directory.glob(
                        f"*.i{(audio_stream or video_stream).itag}*"
                    ),
                    key=extract_sequence_number,
                )[-1]
            )
            sequences_to_download = range(
                latest_sequence_number, rewind_interval.end.sequence + 1
            )
            completed_segments = latest_sequence_number - rewind_interval.start.sequence
        else:
            sequences_to_download = rewind_interval.sequences
            completed_segments = 0

        total_segments = len(rewind_interval.sequences)

        progress_reporter = actions.download.RichProgressReporter()
        if audio_stream:
            progress_reporter.progress.add_task(
                "   - Audio", total=total_segments, completed=completed_segments
            )
        if video_stream:
            progress_reporter.progress.add_task(
                "   - Video", total=total_segments, completed=completed_segments
            )

        do_download_segments = partial(
            actions.download.download_segments,
            playback=playback,
            sequence_numbers=sequences_to_download,
            streams=[x for x in [audio_stream, video_stream] if x is not None],
            output_directory=segments_output_directory,
            progress_reporter=progress_reporter,
        )

        if no_merge:
            click.echo(
                "(<<) Downloading segments {}-{} (no merge requested)...".format(
                    rewind_interval.start.sequence, rewind_interval.end.sequence
                )
            )
            do_download_segments()
            click.echo(
                "\nSuccess! Segments saved to '{}'.".format(
                    try_get_relative_path(segments_output_directory)
                )
            )
        else:
            click.echo("(<<) Preparing and saving the excerpt...")
            click.echo(
                "1. Downloading segments {}-{}:".format(
                    rewind_interval.start.sequence, rewind_interval.end.sequence
                )
            )

            do_download_segments()
            audio_and_video_segment_paths: list[list[Path]] = [[], []]
            if audio_stream:
                audio_and_video_segment_paths[0] = [
                    segments_output_directory
                    / compose_default_segment_filename(sequence, audio_stream.base_url)
                    for sequence in rewind_interval.sequences
                ]
            if video_stream:
                audio_and_video_segment_paths[1] = [
                    segments_output_directory
                    / compose_default_segment_filename(sequence, video_stream.base_url)
                    for sequence in rewind_interval.sequences
                ]

            if cut:
                click.echo("2. Merging segments (cut requested)... ", nl=False)
            else:
                click.echo("2. Merging segments (may take a while)... ", nl=False)

            if not no_metadata:

                def _convert_date_to_isostring(date: datetime) -> str:
                    iso_date = date.astimezone(timezone.utc)
                    return iso_date.strftime("%Y-%m-%dT%H:%M:%S.%fZ")

                metadata_date_converter: Callable[[datetime], str]
                match ctx.obj.config.traverse("output.metadata.dates"):
                    case "unix":
                        metadata_date_converter = lambda d: f"{d.timestamp():.6f}"
                    case "iso":
                        metadata_date_converter = _convert_date_to_isostring
                    case _ as value:
                        click.echo()
                        logger.warning(
                            f"Unknown 'output.metadata.dates = {value}' config value, "
                            "fallback to 'iso'"
                        )
                        metadata_date_converter = _convert_date_to_isostring

                metadata_context: MetadataTagsContext = {
                    "title": playback.info.title,
                    "author": playback.info.author,
                    "comment": playback.video_url,
                    "input_start_date": metadata_date_converter(
                        requested_date_interval.start
                    ),
                    "input_end_date": metadata_date_converter(
                        requested_date_interval.end
                    ),
                    "actual_start_date": metadata_date_converter(
                        actual_date_interval.start
                    ),
                    "actual_end_date": metadata_date_converter(
                        actual_date_interval.end
                    ),
                    "start_segment": str(rewind_interval.start.sequence),
                    "end_segment": str(rewind_interval.end.sequence),
                }

            merged_path = merge_segments(
                audio_and_video_segment_paths[0],
                audio_and_video_segment_paths[1],
                output_directory=final_output_path.parent,
                output_stem=final_output_path.name,
                temp_directory=playback.get_temp_directory(),
                metadata_tags=metadata_context,
                **cut_kwargs,
            )

            click.echo("done.\n")
            click.echo(f"Success! Saved to '{try_get_relative_path(merged_path)}'.")

            if keep_segments and not preview_mode:
                click.echo(
                    "~ Segments are kept in '{}'.".format(
                        try_get_relative_path(segments_output_directory)
                    )
                )

    run_temp_directory = playback.get_temp_directory()
    if keep_temp:
        echo_notice(f"No cleanup enabled, check {run_temp_directory}")
    else:
        try:
            shutil.rmtree(run_temp_directory)
        except OSError:
            logger.warning(
                "Failed to remove temporary directory: %s", run_temp_directory
            )

    if not dry_run and not preview_mode:
        resume_file_path.unlink()

    if not (dry_run or keep_segments or preview_mode):
        try:
            for paths in audio_and_video_segment_paths:
                for path in paths:
                    path.unlink()
        except OSError:
            logger.warning("Could not remove segment: %s", path)
        try:
            if need_to_remove_segments_directory:
                logger.debug(
                    "Remove segments directory: %s",
                    segments_directory_to_remove,
                )
                remove_directories_between(
                    segments_directory_to_remove, segments_output_directory
                )
        except OSError:
            logger.warning(
                "Could not remove segments directory: %s",
                segments_directory_to_remove,
            )



import re
from operator import attrgetter
from typing import Any, Callable, Protocol, Sequence

import click
import lark
import pycond as pc
import structlog

from ytpb.errors import QueryError
from ytpb.types import AudioOrVideoStream

logger = structlog.get_logger(__name__)

pc.ops_use_symbolic_and_txt(allow_single_eq=True)

ALIAS_RE = re.compile(r"@([\w<>=\-\\]+[\?!]?)")

GRAMMAR = r"""
    start: query

    query: expression

    ?expression: ALL -> all
               | NONE -> none
               | conditional_expression
               | fallback_expression
               | piped_expression
               | "(" expression ")" -> group
    conditional_expression: condition
    piped_expression: expression ("|" (expression | function))+ -> pipe
    fallback_expression: expression ("?:" expression)+

    condition: CONDITION_STRING

    function.2: FUNCTION_NAME

    ALL: "all"
    NONE: "none" | "''" | "\"\""
    CONDITION_STRING: /[@a-z0-9_\.,<>=!:\[\]'"\s]+/i
    FUNCTION_NAME: /[a-z0-9_\-]+/i

    %import common.WS_INLINE
    %ignore WS_INLINE
"""


class QueryFunction(Protocol):
    def __call__(
        self, streams: list[AudioOrVideoStream]
    ) -> list[AudioOrVideoStream]: ...


def treat_none_as_false(operator_function: Callable, a: Any, b: Any) -> bool:
    if a is None:
        return False
    else:
        return operator_function(a, b)


def custom_lookup(
    attribute: str, target: str | int, cfg: dict, state: AudioOrVideoStream, **kwargs
) -> tuple[str, str | int]:
    """Custom lookup function for audio and video streams (representations) with
    the existence check of attributes."""
    try:
        if isinstance(attribute, tuple):
            raise QueryError(
                "Attribute incorrectly matched in format spec: '{}'".format(
                    " ".join(attribute)
                )
            )
        if isinstance(state, dict):
            attribute_value = state[attribute]
        else:
            attribute_value = getattr(state, attribute)
    except (AttributeError, KeyError) as e:
        raise QueryError(f"Unknown attribute in format spec: '{attribute}'") from e
    return attribute_value, target


def make_filter_from_expression(expression: str):
    return pc.make_filter(
        expression,
        autoconv_lookups=True,
        lookup=custom_lookup,
        ops_thru=treat_none_as_false,
    )


def parse_expression(value: str) -> lark.Tree:
    return lark.Lark(GRAMMAR, start="query").parse(value)


def execute_query[
    T
](
    tree: lark.Tree,
    items: Sequence[T],
    functions: dict[str, QueryFunction] | None = None,
) -> Sequence[T]:
    def _process_query_tree(tree: lark.Tree, items: Sequence[T]) -> Sequence[T]:
        match tree.data:
            case "all":
                return items
            case "none":
                return []
            case "conditional_expression":
                (condition_node,) = tree.children[0].children
                condition_value = condition_node.value.strip()
                if condition_value == "all":
                    queried = items
                else:
                    condition_filter = make_filter_from_expression(condition_value)
                    queried = list(filter(condition_filter, items))
            case "function":
                assert functions, "No query functions are provided"
                if not items:
                    return items
                function_name = tree.children[0].value
                try:
                    function = functions[function_name]
                except KeyError:
                    raise QueryError(f"Cannot find '{function_name}' query function")
                queried = function(items)
            case "pipe":
                queried = items
                for node in tree.children:
                    queried = _process_query_tree(node, queried)
            case "fallback_expression":
                for node in tree.children:
                    if queried := _process_query_tree(node, items):
                        break
            case "group":
                queried = _process_query_tree(tree.children[0], items)
            case _:
                queried = items
        return queried

    return _process_query_tree(tree, items)


def query_items[
    T
](
    expression: str,
    items: Sequence[T],
    functions: dict[str, QueryFunction] | None = None,
) -> Sequence[T]:
    all_functions = FUNCTIONS | (functions or {})
    tree = parse_expression(expression)
    return execute_query(tree.children[0], items, all_functions)


def best(streams: list[AudioOrVideoStream]) -> list[AudioOrVideoStream]:
    """Gets the best stream in terms of quality (height and frame rate)."""
    return [sorted(streams, key=attrgetter("quality"))[-1]]


def worst(streams: list[AudioOrVideoStream]) -> list[AudioOrVideoStream]:
    """Gets the worst stream in terms of quality (height and frame rate)."""
    return [sorted(streams, key=attrgetter("quality"))[0]]


FUNCTIONS: dict[str, QueryFunction] = {
    "best": best,
    "worst": worst,
}



import enum
import re
import unicodedata
from pathlib import Path

import pathvalidate
import unidecode

from ytpb.utils.other import normalize_info_string


DASHES = [
    "\u002D",  # HYPHEN-MINUS
    "\u2010",  # HYPHEN
    "\u2011",  # NON-BREAKING HYPHEN
    "\u2012",  # FIGURE DASH
    "\u2013",  # EN DASH
    "\u2014",  # EM DASH
    "\u2015",  # HORIZONTAL BAR
]


class AllowedCharacters(enum.StrEnum):
    UNICODE = enum.auto()
    ASCII = enum.auto()
    POSIX = enum.auto()


def sanitize_for_filename(value: str, replacement: str = "-") -> str:
    chars_to_replace = ("|", "/")
    for char in chars_to_replace:
        value = value.replace(char, replacement)
    normalized = unicodedata.normalize("NFKC", value)
    return pathvalidate.sanitize_filename(normalized)


def sanitize_filepath(value: Path) -> Path:
    return pathvalidate.sanitize_filepath(value)


def posixify_for_filename(value: str, separator: str = "-"):
    posix_characters_re = re.compile(r"[^\w\-.]+", flags=re.ASCII)

    if not posix_characters_re.match(separator):
        actual_separator = separator
    else:
        actual_separator = "-"

    output = unidecode.unidecode(value, "ignore")
    output = re.sub(r"(?:\s+)?(\-+)(?:\s+)?", r"\1", output)
    output = posix_characters_re.sub(actual_separator, output)

    if output != actual_separator:
        output = output.strip(actual_separator)

    return output


def adjust_for_filename(
    value: str,
    characters: AllowedCharacters = AllowedCharacters.UNICODE,
    length: int = 255,
    separator: str = " ",
    break_words: bool = False,
) -> str:
    fallback_separator = "-"
    dashes_pattern = r"(?:\s+)?([{0}]+)(?:\s+)?".format("".join(DASHES))

    sanitized = sanitize_for_filename(value, fallback_separator)
    output = normalize_info_string(sanitized)

    is_separator_non_empty = bool(separator and not separator.isspace())
    if is_separator_non_empty or characters == AllowedCharacters.POSIX:
        # For visual appeal, replace dashes and surrounding spaces with dashes
        # to replace them afterward (in parentheses) with a separator
        # symbol. For example: 'A B - C D' -> 'A_B-C_D' (-> 'A_B_C_D').
        output = re.sub(dashes_pattern, r"\1", output)

    actual_separator: str
    if characters == AllowedCharacters.POSIX:
        actual_separator = posixify_for_filename(separator or fallback_separator)
        output = posixify_for_filename(output, actual_separator)
    else:
        match characters:
            case AllowedCharacters.UNICODE:
                actual_separator = separator
            case AllowedCharacters.ASCII:
                actual_separator = separator and unidecode.unidecode(
                    separator, "replace", fallback_separator
                )
                output = unidecode.unidecode(output, "ignore")

    # For the POSIX case, the output is already with spaces replaced.
    if separator != " " and characters != AllowedCharacters.POSIX:
        output = re.sub(dashes_pattern, r"\1", output)
        output = output.replace(" ", actual_separator)

    if length and len(output) > length:
        truncated = output[:length]
        if break_words:
            output = truncated
        else:
            output = truncated.rsplit(actual_separator, 1)[0]
        output = re.sub(dashes_pattern + "$", "", output)

    output = output.strip(fallback_separator + " ")

    return output


def try_get_relative_path(path: Path, other: Path | None = None) -> Path:
    try:
        return path.relative_to(other or Path.cwd())
    except ValueError:
        return path


def remove_directories_between(top: Path, until: Path) -> None:
    until.rmdir()
    directory = until.resolve()
    while directory != top.resolve():
        directory = directory.parent
        directory.rmdir()



""" Fixtures for the CLI tests."""

import contextlib
import os
import sys
from collections.abc import Callable, Sequence
from functools import partial
from pathlib import Path

import pytest
from click import Command
from click.testing import CliRunner, Result

from ytpb.cli import cli


class CustomCliRunner(CliRunner):
    def invoke(
        self, cli: Command, args: str | Sequence[str] | None = None, **kwargs
    ) -> Result:
        sys.argv = args
        return super().invoke(cli, args, **kwargs)


@contextlib.contextmanager
def isolated_filesystem(path: Path):
    old_cwd = Path.cwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_cwd)


@pytest.fixture()
def ytpb_cli_invoke(tmp_path: Path) -> Callable:
    runner = CustomCliRunner()
    with isolated_filesystem(tmp_path):
        yield partial(runner.invoke, cli)



"""Fetchers are used to gather essential information about videos.

Such information includes the basic video information and streams.
"""

from abc import ABC, abstractmethod

import requests
import structlog
from yt_dlp import DownloadError, YoutubeDL

from ytpb.errors import BroadcastStatusError
from ytpb.info import BroadcastStatus, extract_video_info, YouTubeVideoInfo
from ytpb.representations import extract_representations
from ytpb.streams import Streams
from ytpb.types import AudioOrVideoStream, AudioStream, SetOfStreams, VideoStream
from ytpb.utils.url import extract_parameter_from_url

logger = structlog.get_logger(__name__)


class InfoFetcher(ABC):
    """A base abstract class for fetchers.

    Notes:
        Keep in mind that fetchers serve to fetch information, not to
        store.

        A good practice would be to treat each fetch operation as atomic and
        invoke it sequentially. This provides flexibility and will avoid
        repeated calls triggered from outside.
    """

    def __init__(self, video_url: str, session: requests.Session | None = None) -> None:
        """Constructs an object of this class.

        Args:
            video_url: A video URL.
            base_url: A segment base URL.
            session: A :class:`requests.Session` object.
        """
        self.video_url = video_url
        self.session = session or requests.Session()

    @abstractmethod
    def fetch_video_info(self):
        """Fetches basic information about a video."""
        raise NotImplementedError

    @abstractmethod
    def fetch_streams(self):
        """Fetches streams available for a video."""
        raise NotImplementedError


class YtpbInfoFetcher(InfoFetcher):
    """A native fetcher."""

    def fetch_video_info(self) -> YouTubeVideoInfo:
        logger.debug("Fetching index webpage and extracting info")

        response = self.session.get(self.video_url)
        response.raise_for_status()

        info = extract_video_info(self.video_url, response.text)
        if info.status != BroadcastStatus.ACTIVE:
            raise BroadcastStatusError("Stream is not live", info.status)
        self._info = info

        return info

    def fetch_streams(self) -> SetOfStreams:
        assert self._info, "Video info is not set"

        logger.debug("Fetching manifest file and extracting streams")

        dash_manifest_url = self._info.dash_manifest_url
        response = self.session.get(dash_manifest_url)
        response.raise_for_status()

        streams_list = extract_representations(response.text)
        streams = Streams(streams_list)

        return streams


class YoutubeDLInfoFetcher(InfoFetcher):
    """A fetcher that uses :mod:`yt_dlp` to gather information.

    All information is extracted from the :class:`~yt_dlp.YoutubeDL`'s
    information dictionary.
    """

    #: The default options passed to :class:`yt_dlp.YoutubeDL`.
    default_options = {
        "live_from_start": True,
        "quiet": True,
    }

    def __init__(
        self,
        video_url: str,
        session: requests.Session | None = None,
        options: dict | None = None,
    ) -> None:
        """Constructs an object of this class.

        Args:
            video_url: A video URL.
            base_url: A segment base URL.
            session: A :class:`requests.Session` object.
            options: Options passed to :class:`yt_dlp.YoutubeDL`.
        """
        super().__init__(video_url, session)
        self._ydl = YoutubeDL(self.default_options | (options or {}))
        self._formats: list[dict] = []

    def fetch_video_info(self) -> YouTubeVideoInfo:
        try:
            extracted = self._ydl.extract_info(self.video_url, download=False)
        except DownloadError as exc:
            raise AssertionError from exc

        try:
            self._formats = extracted["formats"]

            match extracted["live_status"]:
                case "is_live":
                    status = BroadcastStatus.ACTIVE
                case "was_live" | "post_live":
                    status = BroadcastStatus.COMPLETED
                case "is_upcoming":
                    status = BroadcastStatus.UPCOMING
                case _:
                    status = BroadcastStatus.NONE
            if status != BroadcastStatus.ACTIVE:
                raise BroadcastStatusError("Stream is not live", status)

            info = YouTubeVideoInfo(
                url=extracted["webpage_url"],
                title=extracted["title"],
                author=extracted["uploader"],
                status=status,
                dash_manifest_url=extracted["formats"][0]["manifest_url"],
            )
        except KeyError:
            raise KeyError("Failed to parse extracted info")

        return info

    def _parse_format_item(self, item: dict) -> AudioOrVideoStream:
        base_url = item["fragment_base_url"]
        raw_mime_type = extract_parameter_from_url("mime", base_url)
        mime_type = raw_mime_type.replace("%2F", "/")

        attributes = {
            "itag": item["format_id"],
            "base_url": base_url,
            "mime_type": mime_type,
        }
        if item["acodec"] != "none":
            attributes.update(
                {
                    "codecs": item["acodec"],
                    "audio_sampling_rate": item["asr"],
                }
            )
            stream = AudioStream(**attributes)
        else:
            attributes.update(
                {
                    "codecs": item["vcodec"],
                    "width": item["width"],
                    "height": item["height"],
                    "frame_rate": item["fps"],
                }
            )
            stream = VideoStream(**attributes)
        return stream

    def fetch_streams(self) -> SetOfStreams:
        assert self._formats, "Formats are not set"

        streams = Streams()
        for format_item in self._formats:
            try:
                stream = self._parse_format_item(format_item)
                streams.add(stream)
            except KeyError:
                raise AssertionError

        return streams



"""Media segments and their metadata."""

import re
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from pathlib import Path

import av
import structlog

from ytpb.errors import YtpbError
from ytpb.types import SegmentSequence, Timestamp
from ytpb.utils.other import US_TO_S

logger = structlog.get_logger(__name__)


@dataclass
class SegmentMetadata:
    """Represents the YouTube segment metadata.

    All timestamp and time-related values are in seconds.
    """

    sequence_number: SegmentSequence
    ingestion_walltime: Timestamp
    ingestion_uncertainty: float
    target_duration: float
    first_frame_time: Timestamp
    first_frame_uncertainty: float
    stream_duration: float = None
    max_dvr_duration: float = None
    streamable: str | None = None
    encoding_alias: str | None = None


@dataclass
class Segment:
    """A media segment."""

    def __init__(self) -> None:
        self.local_path: Path | None = None
        self.metadata: SegmentMetadata | None = None
        self.sequence: SegmentSequence | None = None
        self.is_partial: bool | None = None

    @classmethod
    def from_file(cls, path: Path) -> "Segment":
        """Creates a :class:`Segment` object by reading file from path."""
        segment = cls()

        with open(path, "rb") as f:
            content = f.read()
            segment.local_path = path

        segment.metadata = Segment.parse_youtube_metadata(content)
        segment.sequence = segment.metadata.sequence_number

        return segment

    @classmethod
    def from_bytes(cls, content: bytes) -> "Segment":
        """Creates a :class:`Segment` object from bytes."""
        segment = cls()
        segment.metadata = Segment.parse_youtube_metadata(content)
        segment.sequence = segment.metadata.sequence_number
        segment.is_partial = True

        return segment

    @property
    def ingestion_start_date(self) -> datetime:
        """A segment ingestion start date.

        Corresponds to the Ingestion-Walltime-Us value.
        """
        timestamp = self.metadata.ingestion_walltime
        return datetime.fromtimestamp(timestamp, timezone.utc)

    @property
    def ingestion_end_date(self) -> datetime:
        """A segment ingestion end date.

        Corresponds to the actual segment duration started from the
        Ingestion-Walltime-Us value.
        """
        return self.ingestion_start_date + timedelta(seconds=self.get_actual_duration())

    @staticmethod
    def parse_youtube_metadata(content: bytes) -> SegmentMetadata:
        """Parses the metadata from the full or partial content of a segment.

        All timestamp and time-related values are converted to units of seconds.

        Notes:
            If partial content is provided, the amount of bytes should be enough
            to cover the metadata header (see
            :attr:`~ytpb.locate.PARTIAL_SEGMENT_SIZE_BYTES`).

        Args:
            content: Full or partial segment byte content.

        Returns:
            A parsed segment metadata.
        """
        optional_fields = (
            "Encoding-Alias",
            "Streamable",
            "Stream-Duration-Us",
            "Max-Dvr-Duration-Us",
        )

        def _search_for_metadata_field(
            name: str, content: bytes, optional: bool = False
        ) -> bytes | None:
            if matched := re.search(rf"{name}:\s(.+)\r\n".encode(), content):
                value = matched.group(1)
            else:
                if not optional:
                    raise YtpbError(f"Failed to parse metadata field: {name}")
                value = None
            return value

        def _convert_to_float_in_s(value: bytes) -> float:
            return float(value.decode()) / (1 / US_TO_S)

        def _convert_to_timestamp_in_s(value: bytes) -> Timestamp:
            return _convert_to_float_in_s(value)

        metadata_fields_map = (
            ("Sequence-Number", lambda x: int(x.decode())),
            ("Ingestion-Walltime-Us", _convert_to_timestamp_in_s),
            ("Ingestion-Uncertainty-Us", _convert_to_float_in_s),
            ("Stream-Duration-Us", _convert_to_float_in_s),
            ("Max-Dvr-Duration-Us", _convert_to_float_in_s),
            ("Target-Duration-Us", _convert_to_float_in_s),
            ("Streamable", lambda x: x.decode()),
            ("First-Frame-Time-Us", _convert_to_timestamp_in_s),
            ("First-Frame-Uncertainty-Us", _convert_to_float_in_s),
            ("Encoding-Alias", lambda x: x.decode()),
        )

        parsed_metadata_fields = {}
        for name, cast_func in metadata_fields_map:
            value_bytes = _search_for_metadata_field(
                name, content, optional=name in optional_fields
            )
            if value_bytes:
                value = cast_func(value_bytes)
                name_as_key = name.removesuffix("-Us").lower().replace("-", "_")
                parsed_metadata_fields[name_as_key] = value

        return SegmentMetadata(**parsed_metadata_fields)

    def get_actual_duration(self) -> float:
        """Gets the actual segment duration in seconds."""
        with av.open(self.local_path) as container:
            first_packet, *_, last_packet = list(container.demux())[:-1]
        end_pts = last_packet.pts + last_packet.duration
        return float((end_pts - first_packet.pts) * first_packet.time_base)



"""Basic information about videos."""

import re
from dataclasses import dataclass
from enum import auto, StrEnum
from typing import Any

from lxml import etree

from ytpb.errors import InfoExtractError
from ytpb.utils.other import normalize_info_string


class BroadcastStatus(StrEnum):
    """Represents a live broadcast status."""

    #: A broadcast is live.
    ACTIVE = auto()
    #: A broadcast has been scheduled but started.
    UPCOMING = auto()
    #: A broadcast has been completed.
    COMPLETED = auto()
    #: A video is not a live broadcast.
    NONE = auto()


@dataclass
class YouTubeVideoInfo:
    """Represents information about a video."""

    #: Youtube video URL.
    url: str
    #: Video's title.
    title: str
    #: Video's author (channel's name).
    author: str
    #: Video' broadcast status.
    status: BroadcastStatus
    dash_manifest_url: str | None = None

    def __post_init__(self):
        self.title = normalize_info_string(self.title)
        self.author = normalize_info_string(self.author)


class LeftNotFetched:
    pass


#: A sentinel object for a value that is intentionally not fetched.
LEFT_NOT_FETCHED = LeftNotFetched()


def _find_one_or_raise(element: etree.Element, path: str, message: str = "") -> Any:
    """Find first subelement or value which matches the given XPath expression
    or raise an InfoExtractError error."""
    try:
        (result,) = element.xpath(path)
        return result
    except ValueError as exc:
        raise InfoExtractError(message) from exc


def _extract_dash_manifest_url(index_page: str) -> str:
    matched = re.search(r"(?<=dashManifestUrl\":\").*?(?=\")", index_page)
    if not matched:
        raise InfoExtractError("Could not find DASH manifest URL")
    return matched[0]


def extract_video_info(url: str, index_page_text: str) -> YouTubeVideoInfo:
    """Extracts an information about a video from an index page.

    Args:
        url: A video URL.
        index_page_text: An index page string content.

    Returns:
        A :class:`YouTubeVideoInfo` filled with extracted attributes.

    Raises:
        InfoExtractError: If failed to extract an attribute.
    """
    index_page_element = etree.HTML(index_page_text)

    # Extracting title and author:
    video_object_element = _find_one_or_raise(
        index_page_element,
        './/*[@itemtype="http://schema.org/VideoObject"]',
        "Could not find a http://schema.org/VideoObject element",
    )
    title = _find_one_or_raise(
        video_object_element,
        './meta[@itemprop="name"]/@content',
        "Could not extract a title",
    )
    author = _find_one_or_raise(
        video_object_element,
        './*[@itemtype="http://schema.org/Person"]/link[@itemprop="name"]/@content',
        "Could not extract an author",
    )

    # Check if a video is a live stream or not:
    if (
        broadcast_event_element := index_page_element.find(
            './/*[@itemtype="http://schema.org/BroadcastEvent"]'
        )
    ) is not None:
        is_completed = bool(broadcast_event_element.find('./meta[@itemprop="endDate"]'))
        if is_completed:
            status = BroadcastStatus.COMPLETED
            dash_manifest_url = None
        else:
            status = BroadcastStatus.ACTIVE
            dash_manifest_url = _extract_dash_manifest_url(index_page_text)
    else:
        status = BroadcastStatus.NONE
        dash_manifest_url = None

    return YouTubeVideoInfo(url, title, author, status, dash_manifest_url)



from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import TypedDict, TypeVar

import jinja2

from ytpb.cli.utils import date, path
from ytpb.types import AudioStream, VideoStream

T = TypeVar("T", str, Path)


class MinimalOutputPathContext(TypedDict):
    #: YouTube video ID.
    id: str
    #: Video's title.
    title: str


class AudioStreamOutputPathContext(TypedDict):
    #: Audio stream (representaiton).
    audio_stream: AudioStream | None


class VideoStreamOutputPathContext(TypedDict):
    #: Video stream (representaiton).
    video_stream: VideoStream | None


class IntervalOutputPathContext(TypedDict):
    #: Input start date.
    input_start_date: datetime
    #: Input end date.
    input_end_date: datetime
    #: Actual start date.
    actual_start_date: datetime
    #: Actual end date.
    actual_end_date: datetime
    #: Actual duration.
    duration: timedelta


def check_is_template(value: str) -> bool:
    return any([delimiter in value for delimiter in ("{#", "{{", "{%")])


def render_template(
    value: T,
    environment: jinja2.Environment,
    context: dict,
) -> T:
    template = environment.from_string(str(value))
    output = template.render(context).strip()
    return type(value)(output)


def render_path_template(
    value: Path, environment: jinja2.Environment, context: dict
) -> Path:
    rendered = render_template(value, environment, context)
    return path.sanitize_filepath(rendered)


def do_adjust_string(
    value: str,
    chars: str = "posix",
    length: int = 30,
    separator: str = " ",
    break_words: bool = False,
) -> str:
    """Adjusts a string for platform-independent filename.

    The default allowed character set is POSIX-compliant (``chars='posix'``) with
    '-' used as the fallback symbol for separator. For other character sets
    (``'ascii'`` or ``'unicode'``), a whitespace separator will be used.

    The filter does the following:

    * Sanitize a string by removing non-valid characters
    * Translate characters to allowed ones (ASCII-only or POSIX-compliant [1])
      or keep them as is (``chars='unicode'``)
    * Reduce the length to the provided value. By default, words are truncated
      at boundaries.

    References:
        1. https://www.gnu.org/software/automake/manual/html_node/Limitations-on-File-Names.html

    Examples:
        1. Allow only POSIX-compliant characters (default):

           .. sourcecode:: jinja

              {{ "Vidéo en direct – 24/7"|adjust }}
              "Video-en-direct--24-7"

        2. Truncate to a shorten length and break words:

           .. sourcecode:: jinja

              {{ "Vidéo en direct – 24/7"|adjust(length=12, break_words=True) }}
              "Video-en-dir"

        3. Use different separator between words:

           .. sourcecode:: jinja

              {{ "Vidéo en direct – 24/7"|adjust(separator='_') }}
              "Video_en_direct--24-7"

        4. Allow only ASCII characters:

           .. sourcecode:: jinja

              {{ "Vidéo en direct – 24/7"|adjust('ascii') }}
              "Video en direct -- 24-7"

        5. Keep original (sanitized) characters and length:

           .. sourcecode:: jinja

              {{ "Vidéo en direct – 24/7"|adjust('unicode', length=255) }}
              "Vidéo en direct – 24-7"
    """
    characters = path.AllowedCharacters[chars.upper()]
    return path.adjust_for_filename(value, characters, length, separator, break_words)


def do_format_iso_date(value: datetime, styles: str = "basic,complete,hh") -> str:
    """Formats a date according to the ISO 8601 standard.

    Supports complete and reduced date and time representations in basic and
    extended formats. The list of available styles: ``basic`` or ``extended``,
    ``complete`` or ``reduced``, ``hh`` or ``hhmm``, ``z``. Can be used together
    separated by a comma.

    If no styles are provided, or not all styles are specified, the default ones
    will be applied separately: ``basic``, ``complete``, ``hh``.

    Examples:
        .. sourcecode:: jinja

           {# Complete representation, basic format #}
           {{ input_start_date|isodate }}
           "20240102T102000+00"

           {# Complete representation of the Zulu time, extended format #}
           {{ input_start_date|isodate('extended,z') }}
           "2024-01-02T10:20:00Z"

           {# Reduced representation, basic format #}
           {{ input_start_date|isodate('reduced') }}
           "20240102T1020+00"

           {# Complete representation, basic format, HHMM offset #}
           {{ input_start_date|isodate('hhmm') }}
           "20240102T102000+0000"
    """
    style_parameters = date.build_style_parameters_from_spec(styles)
    return date.format_iso_datetime(value, style_parameters)  # type: ignore


def do_convert_to_utc(value: datetime) -> str:
    """Converts a date to the UTC timezone.

    Example:
        .. sourcecode:: jinja

           {{ input_start_date|utc }}
           "2024-01-02 10:20:30.123456+00:00"
    """
    return value.astimezone(timezone.utc)


def do_convert_to_timestamp(value: datetime) -> int:
    """Converts a date to an Unix timestamp in seconds.

    Example:
        .. sourcecode:: jinja

            {{ input_start_date|timestamp }}
            1704190830
    """
    return int(value.timestamp())


def do_format_duration(value: timedelta, style: str = "iso") -> str:
    """Formats a timedelta to a duration string.

    Available styles: ``hms``, ``iso`` (default), ``numeric``.

    Examples:
        .. sourcecode:: jinja

           {{ duration|duration }}
           "PT1H20M30S"

           {{ duration|duration('hms') }}
           "1h20m30s"

           {{ duration|duration('numeric') }}
           "01:20:30"
    """
    pattern = date.DurationFormatPattern[style.upper()]
    return date.format_duration(value, pattern)


FILTERS = {
    "adjust": do_adjust_string,
    "utc": do_convert_to_utc,
    "isodate": do_format_iso_date,
    "timestamp": do_convert_to_timestamp,
    "duration": do_format_duration,
}

# Aliases for Sphinx (autodoc) documentation:
adjust = FILTERS["adjust"]
isodate = FILTERS["isodate"]
utc = FILTERS["utc"]
timestamp = FILTERS["timestamp"]
duration = FILTERS["duration"]



"""Actions to capture frames as images."""

from collections.abc import Iterator
from datetime import datetime

import av
import structlog
from PIL import Image

from ytpb.locate import SegmentLocator
from ytpb.playback import Playback
from ytpb.segment import Segment
from ytpb.types import AudioOrVideoStream, SegmentSequence

logger = structlog.get_logger(__name__)


def extract_frame_as_image(
    segment: Segment, target_date: datetime, last_as_fallback: bool = True
) -> Image.Image:
    """Extracts a frame as image from a segment.

    Args:
        segment: A :class:`~ytpb.segment.Segment` object.
        target_date: A frame target date.
        last_as_fallback: Whether to use a last frame as a fallback if the
          target date is out of a video.

    Returns:
        An extracted image as a :class:`PIL.Image.Image` object.
    """
    target_offset = (target_date - segment.ingestion_start_date).total_seconds()
    with av.open(str(segment.local_path)) as container:
        stream = container.streams.video[0]
        target_pts = stream.start_time + target_offset / stream.time_base
        previous_frame: av.VideoFrame = None
        for current_frame in container.decode(stream):
            if current_frame.pts >= target_pts:
                break
            previous_frame = current_frame
        else:
            if last_as_fallback:
                message = "Target date is out of the video"
                logger.debug(f"{message}, use last frame", date=target_date)
            else:
                raise ValueError(message)
    return (previous_frame or current_frame).to_image()


def capture_frames(
    playback: Playback,
    target_dates: list[datetime],
    stream: AudioOrVideoStream,
    reference_sequence: SegmentSequence | None = None,
) -> Iterator[tuple[Image.Image, Segment]]:
    """Captures frames as images.

    Examples:
        Here's an example of capturing frames from one day every hour::

          from datetime import datetime, timedelta, timezone

          best_stream, = playback.streams.query(
              "type eq video and format eq mp4 and frame_rate eq 30 | best"
          )
          start_date = datetime(2024, 1, 2, 0, tzinfo=timezone.utc)
          dates_to_capture = [start_date + timedelta(hours=h) for h in range(25)]

          captured = capture_frames(
              playback, dates_to_capture, best_stream
          )
          for i, (image, _) in enumerate(captured):
              image.save(f"output-{i:02d}.jpg", quality=80)

    Args:
        playback: A :class:`~ytpb.playback.Playback` object.
        target_dates: A list of dates to capture.
        stream: A stream to which segments used for capturing belongs.
        reference_sequence: A segment sequence number used as a start reference.

    Returns:
        An iterator with pairs of a captured frame and corresponding segment.
    """
    number_of_targets = len(target_dates)
    previous_sequence = reference_sequence
    for i, target_date in enumerate(target_dates):
        sl = SegmentLocator(
            stream.base_url,
            reference_sequence=previous_sequence,
            temp_directory=playback.get_temp_directory(),
            session=playback.session,
        )
        is_end = i == number_of_targets - 1
        found_sequence, *_ = sl.find_sequence_by_time(target_date.timestamp(), is_end)
        previous_sequence = found_sequence

        segment = playback.get_segment(found_sequence, stream)
        image = extract_frame_as_image(segment, target_date)

        yield image, segment



"""Actions to download excerpts."""

from itertools import repeat
from pathlib import Path
from typing import Any, Callable, Iterable, NamedTuple, Protocol, Union

import structlog
from rich.console import Console
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    Progress,
    TaskID,
    TaskProgressColumn,
    TextColumn,
    TimeRemainingColumn,
)

from ytpb.download import (
    compose_default_segment_filename,
    iter_segments,
    save_segment_to_file,
)
from ytpb.merge import merge_segments
from ytpb.playback import Playback, RewindInterval
from ytpb.types import (
    AudioOrVideoStream,
    AudioStream,
    SegmentSequence,
    SetOfStreams,
    VideoStream,
)

logger = structlog.get_logger(__name__)


class ProgressReporter(Protocol):
    def __enter__(self): ...

    def __exit__(self, exc_type, exc_value, traceback): ...

    def update(self, task: int): ...


class NullProgressReporter:
    def __enter__(self) -> None:
        pass

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        pass

    def update(self, task: int = 0) -> None:
        pass


class RichProgressReporter:
    def __init__(self, progress: Progress | None = None):
        if progress is None:
            progress = Progress(
                TextColumn("{task.description}"),
                BarColumn(bar_width=28),
                MofNCompleteColumn(),
                TaskProgressColumn(),
                TextColumn("eta"),
                TimeRemainingColumn(),
                console=Console(),
            )
        self.progress = progress

    def __enter__(self) -> None:
        self.progress.start()

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.progress.stop()

    def update(self, task: int = 0) -> None:
        self.progress.advance(TaskID(task))


class ExcerptDownloadResult(NamedTuple):
    """Represents an excerpt download result."""

    #: An exception raised during downloading.
    exception: Exception | None
    #: A path of the merged excerpt file.
    merged_path: Path
    #: Paths of downloaded audio segments.
    audio_segment_paths: list[Path]
    #: Paths of downloaded video segments.
    video_segment_paths: list[Path]


def chained_zip(*iterables) -> Iterable[Any]:
    """Makes an iterator that returns elements from iterables in round robin fashion.

    Notes:
        The implementation assumes that the lengths of the iterables are the
        same. Otherwise, the iteration will stop after exhausting of the shortest
        iterable.

    Examples:
        >>> list(chained_zip(["A", "B"], [0, 1]))
        ["A", 0, "B", 1]
    """
    iterators = [iter(x) for x in iterables]
    while True:
        for it in iterators:
            try:
                yield next(it)
            except StopIteration:
                return


def download_segments(
    playback: Playback,
    sequence_numbers: Iterable[SegmentSequence],
    streams: Union[SetOfStreams, list[AudioOrVideoStream]],
    output_directory: Path | None = None,
    output_filename: Callable[
        [SegmentSequence, str], str
    ] = compose_default_segment_filename,
    progress_reporter: ProgressReporter | None = None,
) -> list[Path]:
    """Downloads segments.

    Args:
        playback: A playback.
        sequence_numbers: Segment sequence numbers to rewind.
        streams: Streams to download.
        output_directory: A directory where to save downloaded segments.
        output_filename: A callable to compose segment filenames.
        progress_reporter: An instance of :class:`ProgressReporter`-like class
          to show downloading progress. Defaults to dummy progress reporter.

    Returns:
        A list of downloaded segment paths.
    """
    if progress_reporter is None:
        progress_reporter = NullProgressReporter()

    if output_directory is None:
        output_directory = playback.locations["segments"]
        output_directory.mkdir(parents=True, exist_ok=True)

    base_urls: list[str] = [s.base_url for s in streams]
    download_generator = chained_zip(
        *[
            zip(
                iter_segments(sequence_numbers, base_url, session=playback.session),
                repeat(task, len(sequence_numbers)),
            )
            for task, base_url in enumerate(base_urls)
        ]
    )

    downloaded_paths: list[list[Path]] = [[] for _ in base_urls]
    with progress_reporter:
        for (response, sequence, base_url), task in download_generator:
            downloaded_path = save_segment_to_file(
                response.content, sequence, base_url, output_directory, output_filename
            )
            downloaded_paths[task].append(downloaded_path)
            progress_reporter.update(task)

    return downloaded_paths


def download_excerpt(
    playback: Playback,
    rewind_interval: RewindInterval,
    output_stem: str | Path,
    segments_directory: Path,
    audio_stream: AudioStream | None = None,
    video_stream: VideoStream | None = None,
    need_cut: bool = True,
    merge_kwargs: dict[str, Any] | None = None,
    progress_reporter: ProgressReporter | None = None,
) -> ExcerptDownloadResult:
    """Downloads and merges audio and/or video segments.

    Notes:
        Downloaded segments are not cleaned up after merging.

    Args:
        playback: A playback.
        rewind_interval: A rewind interval.
        output_stem: A full path stem of the merged excerpt file.
        audio_stream: An audio stream.
        video_stream: A video stream.
        segments_directory: A directory where to store downloaded segments.
        need_cut: Whether to cut boundary segments to exact times.
        merge_kwargs: Arguments that :meth:`ytpb.merge.merge_segments` takes.
        progress_reporter: An instance of :class:`ProgressReporter`-like class
          to show downloading progress. Defaults to dummy progress reporter.

    Returns:
        An :class:`ExcerptDownloadResult` object.
    """
    if not (audio_stream or video_stream):
        raise AssertionError("At least audio or video stream should be provided")

    if progress_reporter is None:
        progress_reporter = NullProgressReporter()

    all_streams = [x for x in [audio_stream, video_stream] if x is not None]

    segments_directory.mkdir(parents=True, exist_ok=True)

    _downloaded_paths: list[list[Path]] = download_segments(
        playback, rewind_interval.sequences, all_streams, segments_directory
    )
    downloaded_paths: list[list[Path]] = [[], []]
    if audio_stream:
        downloaded_paths[0] = _downloaded_paths[0]
    if video_stream:
        downloaded_paths[1] = _downloaded_paths[1]

    cut_at_kwargs: dict[str, float] = {}
    if need_cut:
        cut_at_kwargs = {
            "cut_at_start": rewind_interval.start.cut_at,
            "cut_at_end": rewind_interval.end.cut_at,
        }

    merge_kwargs = {**cut_at_kwargs, **(merge_kwargs or {})}

    output_stem = Path(output_stem)
    merge_output_directory = output_stem.parent
    merge_output_stem = output_stem.name

    downloaded_audio_paths, downloaded_video_paths = downloaded_paths
    try:
        merged_path = merge_segments(
            downloaded_audio_paths,
            downloaded_video_paths,
            output_directory=merge_output_directory,
            output_stem=merge_output_stem,
            temp_directory=playback.get_temp_directory(),
            **merge_kwargs,
        )
        result = ExcerptDownloadResult(
            None, merged_path, downloaded_audio_paths, downloaded_video_paths
        )
    except Exception as exc:
        result = ExcerptDownloadResult(
            exc, None, downloaded_audio_paths, downloaded_video_paths
        )

    return result



from typing import Any

import pytest

from ytpb.format_spec import query_items


class TestFormatSpecQuery:
    state = [
        {"a": 1},
        {"a": 2},
        {"a": 3},
    ]
    functions = {"first": lambda items: [items[0]] if len(items) > 0 else []}

    @pytest.mark.parametrize(
        "query,expected",
        [
            # Conditional expressions:
            ("a eq 1", [{"a": 1}]),
            ("a = 1", [{"a": 1}]),
            ("a <= 1", [{"a": 1}]),
            ("[a eq 1]", [{"a": 1}]),
            # Piped expressions:
            ("a le 2 | a le 1", [{"a": 1}]),
            ("a le 2 | a le 1 | a eq 1", [{"a": 1}]),
            ("a le 2 | a eq 0 | a eq 1", []),
            # Fallback expressions:
            ("a eq 1 ?: a eq 2", [{"a": 1}]),
            ("a eq 0 ?: a eq 2", [{"a": 2}]),
            ("a eq 0 ?: a eq 100", []),
            # Grouped expressions:
            ("(a eq 1)", [{"a": 1}]),
            ("(a le 2 | a le 1) | a eq 1", [{"a": 1}]),
            ("(a eq 0 ?: a eq 100) ?: a eq 1", [{"a": 1}]),
            # Expressions with function:
            ("a le 2 | first", [{"a": 1}]),
            ("a eq 0 | first", []),
            # Mixed expressions:
            ("a le 2 ?: a le 3 | first", [{"a": 1}, {"a": 2}]),
            ("a eq 0 | first ?: a le 2 ", [{"a": 1}, {"a": 2}]),
            ("a eq 0 | first ?: a le 2 | first", [{"a": 1}]),
            ("a eq 0 | first ?: a le 3 | a le 2 | first", [{"a": 1}]),
            ("a eq 0 | first ?: a le 3 | first | first", [{"a": 1}]),
            ("(a le 2 | first) ?: a le 3", [{"a": 1}]),
            ("(a eq 2 | first) ?: (a le 3 | first)", [{"a": 2}]),
            ("(a eq 0 | first) ?: (a le 3 | first)", [{"a": 1}]),
            # Others:
            ("all", [{"a": 1}, {"a": 2}, {"a": 3}]),
            ("none", []),
            ("''", []),
            ('""', []),
        ],
    )
    def test_query_expression(self, query: str, expected: list[dict[str, Any]]):
        assert expected == query_items(query, self.state, self.functions)



import re
from operator import attrgetter
from typing import Any, Callable, Protocol, Sequence

import click
import lark
import pycond as pc
import structlog

logger = structlog.get_logger(__name__)


ALIAS_RE = re.compile(r"@([\w<>=\-\\]+[\?!]?)")


def expand_aliases(expression: str, aliases: dict[str, str]) -> str:
    source_alias: str = ""
    expanding_branch: set[str] = set()

    def _resolve_aliases(
        expression: str,
        aliases: dict[str, str],
        patterns: list[tuple[str, str]],
    ) -> str:
        nonlocal source_alias
        nonlocal expanding_branch

        resolved = expression
        for matched in ALIAS_RE.finditer(expression):
            current_alias = matched.group()
            current_name = matched.group(1)
            if current_name in expanding_branch:
                raise click.UsageError(
                    f"Cannot resolve circular alias {current_alias} in {source_alias}"
                )
            else:
                expanding_branch.add(current_name)
                source_alias = current_alias
            try:
                value = aliases[current_name]
                if not ALIAS_RE.search(value):
                    expanding_branch = set()
            except KeyError:
                value = current_name
                for pattern, repl in patterns:
                    value, has_subbed = re.subn(rf"^{pattern}", repl, value)
                    if has_subbed:
                        break
                else:
                    raise click.UsageError(f"Unknown alias found: {current_alias}")
            resolved = resolved.replace(current_alias, value)

        if ALIAS_RE.search(resolved):
            return _resolve_aliases(resolved, aliases, patterns)

        return resolved

    patterns = [(k, v) for k, v in aliases.items() if "\\" in k]
    output = _resolve_aliases(expression, aliases, patterns)

    return output


# Built-in aliases

MEDIA_FORMAT_ALIASES = {"mp4": "format = mp4", "webm": "format = webm"}

CODECS_ALIASES = {
    "mp4a": "codecs contains mp4a",
    "avc1": "codecs contains avc1",
    "vp9": "codecs = vp9",
}

NAMED_QUALITY_ALIASES = {
    "low": "height = 144",
    "medium": "height = 480",
    "high": "height = 720",
    "FHD": "height = 1080",
    "2K": "height = 1440",
    "4K": "height = 2160",
}

ITAG_PATTERN_ALIAS = (r"\b(\d+)\b", r"itag = \1")
QUALITY_PATTERN_ALIAS = (r"(\d+)p\b", r"height = \1")
QUALITY_FPS_PATTERN_ALIAS = (r"(\d+)p(\d+)\b", r"[height = \1 and frame_rate = \2]")
QUALITY_OP_PATTERN_ALIAS = (r"([<>=]=?)(\d+)p\b", r"height \1 \2")
FPS_PATTERN_ALIAS = (r"(\d+)fps\b", r"frame_rate = \1")

PATTERN_ALIASES = dict(
    (
        ITAG_PATTERN_ALIAS,
        QUALITY_PATTERN_ALIAS,
        QUALITY_FPS_PATTERN_ALIAS,
        QUALITY_OP_PATTERN_ALIAS,
        FPS_PATTERN_ALIAS,
    ),
)

ALIASES: dict[str, str] = {
    **MEDIA_FORMAT_ALIASES,
    **NAMED_QUALITY_ALIASES,
    **CODECS_ALIASES,
    **PATTERN_ALIASES,
}



from __future__ import annotations

from typing import TYPE_CHECKING

import requests

if TYPE_CHECKING:
    from ytpb.info import BroadcastStatus


class YtpbError(Exception):
    """Base Ytpb exception."""


class BroadcastStatusError(YtpbError):
    """Raised because of :class:`ytpb.info.BroadcastStatus`."""

    def __init__(self, message: str, status: BroadcastStatus):
        self.status = status
        super().__init__(message)


class InfoExtractError(YtpbError):
    """Failed to extract a YouTube video information."""


class BaseUrlExpiredError(YtpbError):
    pass


class CachedItemNotFoundError(YtpbError):
    """Failed to find an unexpired cached item."""

    def __str__(self) -> str:
        return "Unexpired cached item doesn't exist for the video"


class MaxRetryError(YtpbError):
    """Raised when retry limit has been exceeded."""

    def __init__(self, message, response):
        self.response = response
        super().__init__(message)


class FFmpegRunError(YtpbError):
    """Raised during FFmpeg subprocess call."""


class QueryError(YtpbError):
    """Failed to query streams with format spec."""


class BadCommandArgument(YtpbError):
    pass


class SequenceLocatingError(YtpbError):
    """Failed to locate a segment."""


class SegmentDownloadError(YtpbError, requests.exceptions.HTTPError):
    """Failed to download a segment."""

    def __init__(self, message: str, sequence: int):
        self.sequence = sequence
        super().__init__(message)



"""Representations from MPEG-DASH MPD."""

from dataclasses import dataclass
from functools import total_ordering

from lxml import etree


NAMESPACES = {
    "mpd": "urn:mpeg:DASH:schema:MPD:2011",
    "yt": "http://youtube.com/yt/2012/10/10",
}


@total_ordering
@dataclass(frozen=True)
class VideoQuality:
    """Represents a video quality (height and frame rate).

    Examples:
        The class supports comparison operations. For example: ::

          >>> VideoQuality(720) == VideoQuality(720, 30)
          True
          >>> VideoQuality(720) > VideoQuality(720, 60)
          False
    """

    height: int
    frame_rate: float

    @classmethod
    def from_string(cls, value: str) -> "VideoQuality":
        """Creates a :class:`VideoQuality` object from string.

        Args:
            value: A video quality string. For example: "720p", "1080p60".
        """
        try:
            height, frame_rate = value.split("p")
            return cls(int(height), float(frame_rate or 30))
        except ValueError:
            raise ValueError("Value not formatted as video quality")

    def __str__(self) -> str:
        if self.frame_rate == 30:
            return f"{self.height}p"
        else:
            return f"{self.height}p{self.frame_rate:.2g}"

    def __eq__(self, other) -> bool:
        return str(self) == str(other)

    def __gt__(self, other) -> bool:
        if isinstance(other, str):
            other = type(self).from_string(other)
        if self.height > other.height:
            return True
        elif self.height == other.height and self.frame_rate > other.frame_rate:
            return True
        else:
            return False


@dataclass(frozen=True, slots=True)
class RepresentationInfo:
    """Represents common attributes of audio and video representations."""

    #: itag value, e.g. '140', '247'.
    itag: str
    #: MIME type, e.g. 'audio/mp4', 'video/webm'.
    mime_type: str
    #: Codec name, e.g. 'mp4', 'vp9'.
    codecs: str
    #: Segment base URL.
    base_url: str

    @property
    def type(self) -> str:
        """An alias for a MIME type, e.g. 'audio', 'video'."""
        return self.mime_type.split("/")[0]

    @property
    def format(self) -> str:
        """An alias for a MIME subtype, e.g. 'mp4', 'webm'."""
        return self.mime_type.split("/")[1]

    def __repr__(self):
        return f"{type(self).__name__}(itag={self.itag})"


@dataclass(frozen=True, repr=False)
class AudioRepresentationInfo(RepresentationInfo):
    """Represents attributes of audio representations."""

    #: Sampling rate (in Hz).
    audio_sampling_rate: int


@dataclass(frozen=True, repr=False)
class VideoRepresentationInfo(RepresentationInfo):
    """Represents attributes of video representations."""

    #: Width of frame.
    width: int
    #: Height of frame.
    height: int
    #: Frame per second (FPS).
    frame_rate: int

    @property
    def quality(self) -> VideoQuality:
        """Quality string (resolution and FPS), e.g. '720p', '1080p60'."""
        return VideoQuality(self.height, self.frame_rate)

    @property
    def fps(self) -> int:
        """An alias for ``frame_rate``."""
        return self.frame_rate


def _eval_local_xpath(element: etree.Element, node: str) -> list[etree.Element]:
    return element.xpath(".//*[local-name() = $node]", node=node)


def extract_representations(manifest_content: str) -> list[RepresentationInfo]:
    """Extracts representations from a manifest.

    Args:
        manifest_content: An MPEG-DASH MPD string content.

    Returns:
        A list of :class:`RepresentationInfo` objects.
    """
    representations_info: list[RepresentationInfo] = []

    manifest = etree.fromstring(manifest_content.encode())

    adaptation_sets = manifest.xpath("//mpd:AdaptationSet", namespaces=NAMESPACES)
    for adaptation in adaptation_sets:
        mime_type = adaptation.get("mimeType")
        is_audio = "audio" in mime_type

        representations = _eval_local_xpath(adaptation, "Representation")
        for repr_ in representations:
            itag = repr_.get("id")
            base_repr_kwargs = dict(
                itag=itag,
                mime_type=mime_type,
                codecs=repr_.get("codecs"),
                base_url=_eval_local_xpath(repr_, "BaseURL")[0].text,
            )

            if is_audio:
                info = AudioRepresentationInfo(
                    **base_repr_kwargs,
                    audio_sampling_rate=int(repr_.get("audioSamplingRate")),
                )
            else:
                info = VideoRepresentationInfo(
                    **base_repr_kwargs,
                    width=int(repr_.get("width")),
                    height=int(repr_.get("height")),
                    frame_rate=int(repr_.get("frameRate")),
                )
            representations_info.append(info)

    return representations_info



import atexit
import logging
import os
import re
import sys
import tomllib
from copy import deepcopy
from dataclasses import dataclass, field
from datetime import datetime, timezone
from fileinput import FileInput
from pathlib import Path
from typing import Any, cast, TextIO, TypeAlias

import click
import cloup
import jinja2.sandbox
import structlog

from ytpb._version import __version__
from ytpb.cli.commands.capture import capture_group
from ytpb.cli.commands.download import download_command
from ytpb.cli.commands.mpd import mpd_group
from ytpb.cli.common import suppress_output
from ytpb.cli.config import (
    AddressableChainMap,
    DEFAULT_CONFIG,
    get_default_config_path,
    load_config_from_file,
    setup_logging,
    update_nested_dict,
)
from ytpb.cli.options import config_options, logging_options
from ytpb.cli.templating import FILTERS as TEMPLATE_FILTERS

ConfigMap: TypeAlias = AddressableChainMap

logger = structlog.get_logger(__name__)


@dataclass
class ReportStreamWrapper:
    stream: TextIO
    file: TextIO

    def write(self, message: str) -> None:
        self.stream.write(message)
        self.file.write(message)

    def __getattr__(self, attr: str) -> Any:
        return getattr(self.stream, attr)


@dataclass
class ContextObject:
    """This object is referenced as `ctx.obj`."""

    config: ConfigMap = field(default_factory=lambda: ConfigMap(DEFAULT_CONFIG))
    original_stdout: TextIO = field(default_factory=lambda: sys.stdout)


def find_option_name_by_flag(command: click.Command, flag: str) -> str:
    for option in command.params:
        for opt in option.opts:
            if flag == opt.lstrip("-"):
                return option.name
    raise click.UsageError(f"Cannot find option with flag '--{flag}'.")


def load_config_into_context(ctx: click.Context, path: Path) -> dict:
    try:
        config_dict = load_config_from_file(path)
    except FileNotFoundError as e:
        raise click.FileError(
            filename=path,
            hint=str(e),
        )
    except tomllib.TOMLDecodeError as e:
        raise click.FileError(
            filename=path,
            hint=f"Could not load configuration file.\n{e}",
        )

    ctx.ensure_object(ContextObject)
    ctx.obj.config = ctx.obj.config.new_child(config_dict)

    # FIXME: This all looks like one unbeatiful hack. Really hope there's
    # another way that doesn't require much effort. See also pallets/click#680.
    try:
        user_options = ctx.obj.config.maps[0]["options"]
    except KeyError:
        pass
    else:
        _invoked_subcommand = ctx.command.commands[ctx.invoked_subcommand]
        if isinstance(_invoked_subcommand, click.Group):
            leftover_args = sys.argv[sys.argv.index(_invoked_subcommand.name) + 1 :]
            deepest_name = [arg for arg in leftover_args if not arg.startswith("-")][0]
            deepest_command = _invoked_subcommand.commands[deepest_name]
            user_defaults = user_options[_invoked_subcommand.name][deepest_name]
        else:
            deepest_command = _invoked_subcommand
            user_defaults = user_options[deepest_command.name]

        user_defaults_renamed: dict[str, str] = {}
        for flag, value in user_defaults.items():
            option_name = find_option_name_by_flag(deepest_command, flag)
            user_defaults_renamed[option_name] = value

        if isinstance(_invoked_subcommand, click.Group):
            user_options[_invoked_subcommand.name] = {
                deepest_command.name: user_defaults_renamed
            }
        else:
            user_options[deepest_command.name] = user_defaults_renamed

        ctx.default_map = update_nested_dict(ctx.default_map, user_options)


@cloup.group(invoke_without_command=True)
@cloup.option_group(
    "Global options",
    config_options,
    logging_options,
    click.option(
        "-q",
        "--quiet",
        help="Supress all normal output.",
        default=False,
        is_flag=True,
        is_eager=True,
    ),
)
@click.version_option(__version__, "-V", "--version", message="%(version)s")
@click.pass_context
def base_cli(
    ctx: click.Context,
    config_path: Path,
    no_config: bool,
    debug: bool,
    report: bool,
    quiet: bool,
) -> None:
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())
        ctx.exit()

    ctx.ensure_object(ContextObject)
    ctx.default_map = ctx.obj.config["options"]

    if quiet:
        suppress_output()

    if report:
        debug = True
        os.environ["NO_COLOR"] = "1"

        timestamp = datetime.now(tz=timezone.utc).strftime("%Y%m%d-%H%M%S")
        report_file_path = Path.cwd() / f"ytpb-{timestamp}.log"
        report_handle = open(report_file_path, "a", encoding="utf-8")
        sys.stdout = cast(
            TextIO, ReportStreamWrapper(ctx.obj.original_stdout, report_handle)
        )
        sys.stderr = cast(TextIO, ReportStreamWrapper(sys.stderr, report_handle))

        @atexit.register
        def sanitize_report_file() -> None:
            report_handle.close()
            with FileInput(report_handle.name, inplace=True) as fi:
                for line in fi:
                    print(re.sub(r"/ip/([\w.:]+)/", "/ip/0.0.0.0/", line), end="")

    setup_logging(logging.DEBUG if debug else logging.WARNING)

    if not no_config:
        if config_path:
            logger.info("Using configuration from '%s' via --config", config_path)
            load_config_into_context(ctx, config_path)
        else:
            default_config_path = get_default_config_path()
            if default_config_path.exists():
                logger.info("Using configuration from '%s'", default_config_path)
                load_config_into_context(ctx, default_config_path)
    else:
        if config_path:
            raise click.UsageError("Conflicting --config and --no-config options given")

    ctx.obj.jinja_environment = jinja2.sandbox.SandboxedEnvironment()
    ctx.obj.jinja_environment.filters.update(TEMPLATE_FILTERS)


cli = deepcopy(base_cli)
cli.help = "A playback for YouTube live streams"
cli.section(
    "Top-level commands",
    download_command,
    capture_group,
    mpd_group,
)



import os
import platform
from pathlib import Path
from typing import Callable
from unittest.mock import MagicMock, patch
from urllib.parse import urljoin

import platformdirs
import pytest
import toml
from freezegun import freeze_time


@pytest.mark.parametrize(
    "audio_formats,video_formats",
    [
        ("itag eq 140", "itag eq 243 or itag eq 244"),
        ("itag eq 140", "none"),
        ("none", "itag eq 243 or itag eq 244"),
    ],
)
@pytest.mark.expect_suffix(platform.system())
def test_compose_mpd(
    audio_formats,
    video_formats,
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    tmp_path: Path,
    expected_out,
) -> None:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        r"https://.+\.googlevideo\.com/videoplayback/.+/sq/\w+"
    )
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "mpd",
                "compose",
                "--no-cache",
                "--interval",
                "7959120/7959121",
                "-af",
                audio_formats,
                "-vf",
                video_formats,
                stream_url,
            ],
        )
    assert result.exit_code == 0
    assert result.output == expected_out


@pytest.mark.expect_suffix(platform.system())
@freeze_time("2023-03-26T00:00:00+00:00")
def test_compose_mpd_with_no_streams(
    ytpb_cli_invoke: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    video_id: str,
    audio_base_url: str,
    video_base_url: str,
    tmp_path: Path,
    expected_out,
) -> None:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "mpd",
                "compose",
                "--no-cache",
                "--interval",
                "2023-03-25T23:33:55+00:00/2023-03-25T23:33:58+00:00",
                "-af",
                "itag eq 0",
                "-vf",
                "itag eq 0",
                stream_url,
            ],
        )

    assert result.exit_code == 1
    assert result.output == expected_out


def test_compose_mpd_using_yt_dlp(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    video_id: str,
    audio_base_url: str,
    video_base_url: str,
    tmp_path: Path,
) -> None:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"),
    )

    with patch("ytpb.cli.common.YoutubeDLInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "mpd",
                "compose",
                "--yt-dlp",
                "--no-cache",
                "--interval",
                "7959120/7959121",
                "-af",
                "itag eq 140",
                "-vf",
                "",
                stream_url,
            ],
        )

    assert result.exit_code == 0


def test_with_default_config(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    tmp_path: Path,
) -> None:
    # Given:
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        r"https://.+\.googlevideo\.com/videoplayback/.+/sq/\w+"
    )

    config = {
        "options": {
            "mpd": {
                "compose": {
                    "audio-formats": "NON-SENS",
                }
            }
        }
    }
    config_path = platformdirs.user_config_path() / "ytpb/config.toml"
    config_path.parent.mkdir(parents=True)
    with config_path.open("w", encoding="utf-8") as f:
        toml.dump(config, f)

    # When:
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "mpd",
                "compose",
                "--no-cache",
                "--interval",
                "7959120/7959121",
                "-af",
                "itag eq 140",
                "-vf",
                "itag eq 243",
                stream_url,
            ],
        )

    # Then:
    print(result.output)
    assert result.exit_code == 0


def test_compose_to_user_output_path(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    tmp_path: Path,
):
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        r"https://.+\.googlevideo\.com/videoplayback/.+/sq/\w+"
    )
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "mpd",
                "compose",
                "--no-cache",
                "--interval",
                "7959120/7959121",
                "-af",
                "itag eq 140",
                "-vf",
                "itag eq 243",
                "-o",
                "manifest.xml",
                stream_url,
            ],
            catch_exceptions=False,
        )
    assert result.exit_code == 0
    assert os.path.exists("manifest.xml")


def test_compose_to_user_template_output_path(
    ytpb_cli_invoke: Callable,
    add_responses_callback_for_reference_base_url: Callable,
    add_responses_callback_for_segment_urls: Callable,
    fake_info_fetcher: MagicMock,
    stream_url: str,
    video_id: str,
    tmp_path: Path,
):
    add_responses_callback_for_reference_base_url()
    add_responses_callback_for_segment_urls(
        r"https://.+\.googlevideo\.com/videoplayback/.+/sq/\w+"
    )
    with patch("ytpb.cli.common.YtpbInfoFetcher") as mock_fetcher:
        mock_fetcher.return_value = fake_info_fetcher
        result = ytpb_cli_invoke(
            [
                "mpd",
                "compose",
                "--no-cache",
                "--interval",
                "7959120/7959121",
                "-af",
                "itag eq 140",
                "-vf",
                "itag eq 243",
                "-o",
                "{{ id }}.xml",
                stream_url,
            ],
            catch_exceptions=False,
        )
    assert result.exit_code == 0
    assert os.path.exists(f"{video_id}.xml")



from dataclasses import dataclass
from pathlib import Path
from typing import Callable
from urllib.parse import urljoin

from ytpb import actions
from ytpb.playback import Playback, RewindInterval

from tests.helpers import assert_approx_duration


@dataclass
class FakeRewindMoment:
    sequence: int
    cut_at: float = 0


@dataclass
class FakeRewindInterval:
    start: FakeRewindMoment
    end: FakeRewindMoment
    sequences = RewindInterval.sequences


@dataclass
class FakeStream:
    base_url: str


def test_download_audio_segments(
    fake_info_fetcher: "FakeInfoFetcher",
    add_responses_callback_for_segment_urls: Callable,
    stream_url: str,
    audio_base_url: str,
    tmp_path: Path,
) -> None:
    # Given:
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"),
    )

    # When:
    playback = Playback(stream_url, fetcher=fake_info_fetcher)
    playback.fetch_and_set_essential()
    output_paths = actions.download.download_segments(
        playback,
        sequence_numbers=range(7959120, 7959122),
        streams=[FakeStream(audio_base_url)],
        output_directory=tmp_path,
    )

    # Then:
    assert output_paths == [
        [
            tmp_path / "7959120.i140.mp4",
            tmp_path / "7959121.i140.mp4",
        ]
    ]


def test_download_audio_and_video_segments(
    fake_info_fetcher: "FakeInfoFetcher",
    add_responses_callback_for_segment_urls: Callable,
    stream_url: str,
    audio_base_url: str,
    video_base_url: str,
    tmp_path: Path,
) -> None:
    # Given:
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"),
        urljoin(video_base_url, r"sq/\w+"),
    )

    # When:
    playback = Playback(stream_url, fetcher=fake_info_fetcher)
    playback.fetch_and_set_essential()
    output_paths = actions.download.download_segments(
        playback,
        sequence_numbers=range(7959120, 7959122),
        streams=[FakeStream(audio_base_url), FakeStream(video_base_url)],
        output_directory=tmp_path,
    )

    # Then:
    assert output_paths == [
        [
            tmp_path / "7959120.i140.mp4",
            tmp_path / "7959121.i140.mp4",
        ],
        [
            tmp_path / "7959120.i244.webm",
            tmp_path / "7959121.i244.webm",
        ],
    ]


def test_download_audio_excerpt_with_cutting(
    fake_info_fetcher: "FakeInfoFetcher",
    add_responses_callback_for_segment_urls: Callable,
    stream_url: str,
    audio_base_url: str,
    run_temp_directory: Path,
    tmp_path: Path,
) -> None:
    # Given:
    add_responses_callback_for_segment_urls(urljoin(audio_base_url, r"sq/\w+"))

    # When:
    playback = Playback(stream_url, fetcher=fake_info_fetcher)
    playback.fetch_and_set_essential()
    output_result = actions.download.download_excerpt(
        playback,
        rewind_interval=FakeRewindInterval(
            FakeRewindMoment(7959120, cut_at=0.5), FakeRewindMoment(7959121, cut_at=1.5)
        ),
        output_stem=tmp_path / "output",
        audio_stream=FakeStream(audio_base_url),
        segments_directory=tmp_path / "segments",
    )

    # Then:
    assert output_result == (
        None,
        tmp_path / "output.mp4",
        [
            tmp_path / "segments/7959120.i140.mp4",
            tmp_path / "segments/7959121.i140.mp4",
        ],
        [],
    )
    assert_approx_duration(output_result[1], 3)


def test_download_audio_and_video_excerpt_without_cutting(
    fake_info_fetcher: "FakeInfoFetcher",
    add_responses_callback_for_segment_urls: Callable,
    stream_url: str,
    audio_base_url: str,
    video_base_url: str,
    run_temp_directory: Path,
    tmp_path: Path,
) -> None:
    # Given:
    add_responses_callback_for_segment_urls(
        urljoin(audio_base_url, r"sq/\w+"), urljoin(video_base_url, r"sq/\w+")
    )

    # When:
    playback = Playback(stream_url, fetcher=fake_info_fetcher)
    playback.fetch_and_set_essential()
    output_result = actions.download.download_excerpt(
        playback,
        rewind_interval=FakeRewindInterval(
            FakeRewindMoment(7959120, cut_at=0.5),
            FakeRewindMoment(7959121, cut_at=1.5),
        ),
        output_stem=tmp_path / "output",
        audio_stream=FakeStream(audio_base_url),
        video_stream=FakeStream(video_base_url),
        segments_directory=tmp_path / "segments",
        need_cut=False,
    )

    # Then:
    assert output_result == (
        None,
        tmp_path / "output.mkv",
        [
            tmp_path / "segments/7959120.i140.mp4",
            tmp_path / "segments/7959121.i140.mp4",
        ],
        [
            tmp_path / "segments/7959120.i244.webm",
            tmp_path / "segments/7959121.i244.webm",
        ],
    )
    assert_approx_duration(output_result[1], 4)



from ytpb import representations

from .conftest import TEST_DATA_PATH


def test_representation_properties(audio_base_url):
    representation = representations.AudioRepresentationInfo(
        itag="140",
        codecs="mp4a.40.2",
        mime_type="audio/mp4",
        base_url=audio_base_url,
        audio_sampling_rate=44100,
    )
    assert representation.type == "audio"
    assert representation.format == "mp4"


def test_extract_representations_info(audio_base_url: str, video_base_url: str):
    with open(TEST_DATA_PATH / "manifest-1695928670.mpd") as f:
        results = representations.extract_representations(f.read())

    assert (
        representations.AudioRepresentationInfo(
            itag="140",
            codecs="mp4a.40.2",
            mime_type="audio/mp4",
            base_url=audio_base_url,
            audio_sampling_rate=44100,
        )
        in results
    )
    assert (
        representations.VideoRepresentationInfo(
            itag="244",
            codecs="vp9",
            mime_type="video/webm",
            base_url=video_base_url,
            width=854,
            height=480,
            frame_rate=30,
        )
        in results
    )



from ytpb.errors import CachedItemNotFoundError
from ytpb.fetchers import InfoFetcher, YtpbInfoFetcher
from ytpb.playback import Playback
from ytpb.utils.url import normalize_video_url


def get_playback(
    stream_url_or_id: str,
    use_cache: bool = True,
    force_update_cache: bool = False,
    fetcher: InfoFetcher | None = None,
) -> Playback:
    """Gets a playback for a live stream.

    By default, it uses cache to store essential information between runs. The
    default cache directory is ``$XDG_CACHE_HOME/ytpb`` (with a fallback to
    ``~/.cache/ytpb``).

    Args:
        stream_url_or_id: A live stream URL or ID.
        use_cache: Whether to use cache.
        force_update_cache: Whether to force update cache item.
        fetcher: An information :ref:`fetcher <ytpb-fetchers>` to use. By
          default, :class:`~ytpb.fetchers.YtpbInfoFetcher` is used.

    Returns:
        A :class:`~ytpb.playback.Playback` object.

    Raises:
        BroadcastStatusError: If broadcast status of a live stream is not
          :attr:`~ytpb.info.BroadcastStatus.ACTIVE`.
    """
    stream_url = normalize_video_url(stream_url_or_id)

    if fetcher is None:
        fetcher = YtpbInfoFetcher(stream_url)

    need_read_cache = use_cache and not force_update_cache
    if need_read_cache:
        try:
            playback = Playback.from_cache(stream_url, fetcher=fetcher)
        except CachedItemNotFoundError:
            playback = Playback.from_url(
                stream_url, fetcher=fetcher, write_to_cache=True
            )
    else:
        playback = Playback.from_url(
            stream_url, fetcher=fetcher, write_to_cache=force_update_cache
        )

    return playback



"""Merge and cut media segments.

Note:
    The current implementation of segment merging is not optimal in terms of
    disk space, as it requires triple the amount of the total size of segments
    if cutting is requested (by default). Should be changed to use different
    techniques, without the need for intermediate files.
"""

import functools
import os
import shlex
from pathlib import Path
from typing import Any

import av

from ytpb.utils import ffmpeg

__all__ = ("merge_segments", "mux_and_cut_boundary_segment")


DEFAULT_VIDEO_ENCODING_SETTINGS = {
    "h264": "libx264 -crf 18",
    "vp9": "libvpx-vp9 -crf 31 -b:v 0",
    "av1": "libaom-av1 -crf 31",
}


def get_nth_or_none(iterable, n: int) -> Any | None:
    if not iterable:
        return None
    return iterable[n]


def mux_and_cut_boundary_segment(
    audio_segment_path: Path,
    video_segment_path: Path,
    output_path: Path,
    video_codec: str | None = None,
    **cut_kwargs,
) -> None:
    """Muxes and cuts a boundary segment.

    Args:
        audio_segment_path: A path to an audio segment.
        video_segment_path: A path to a video segment.
        output_path: An output path of the muxed segment.
        video_codec: A video codec name.
        cut_kwargs: Cut keyword arguments: ``cut_at_start`` and ``cut_at_end``.
    """

    def prepare_ffmpeg_input_options(
        segment_path: Path, cut_at_start: float = 0, cut_at_end: float = 0
    ):
        assert cut_at_start or cut_at_end
        if cut_at_start > 0:
            return ["-ss", f"{cut_at_start:.6f}s", "-i", segment_path]
        elif cut_at_end > 0:
            return ["-i", segment_path, "-to", f"{cut_at_end:.6f}s"]
        else:
            return ["-i", segment_path]

    if not {"cut_at_start", "cut_at_end"} > cut_kwargs.keys():
        raise ValueError(
            "only cut_at_start or cut_at_end keyword argument is accepted in cut_kwargs"
        )

    if video_segment_path and video_codec is None:
        with av.open(video_segment_path) as container:
            video_av_stream = container.streams.get({"video": 0})[0]
            video_codec = video_av_stream.codec_context.name

    ffmpeg_input_options = []
    ffmpeg_codecs_options = []

    should_cut = bool(
        cut_kwargs.get("cut_at_start", 0) or cut_kwargs.get("cut_at_end", 0)
    )
    if should_cut is False:
        if video_segment_path:
            ffmpeg_input_options += ["-i", video_segment_path]
        if audio_segment_path:
            ffmpeg_input_options += ["-i", audio_segment_path]
        ffmpeg.run_ffmpeg(ffmpeg_input_options + ["-c", "copy", output_path])
    else:
        if video_segment_path:
            ffmpeg_input_options += prepare_ffmpeg_input_options(
                video_segment_path, **cut_kwargs
            )

            user_encoding_settings = os.environ.get(
                f"YTPB_{video_codec.upper()}_ENCODING_SETTINGS", None
            )
            if user_encoding_settings:
                ffmpeg_codecs_options += ["-c:v"] + shlex.split(user_encoding_settings)
            else:
                try:
                    ffmpeg_codecs_options += ["-c:v"] + shlex.split(
                        DEFAULT_VIDEO_ENCODING_SETTINGS[video_codec]
                    )
                except KeyError:
                    raise ValueError(
                        "No encoding settings are availabe for "
                        f"'{video_codec}' video codec"
                    )

        if audio_segment_path:
            ffmpeg_input_options += prepare_ffmpeg_input_options(
                audio_segment_path, **cut_kwargs
            )
            ffmpeg_codecs_options += ["-c:a", "copy"]

        additional_options: list[str] = []
        if video_segment_path and video_codec == "h264":
            # Ensure to have the same tbn values after re-encoding of segments
            # and their concatenation (see the merge_segments() function).
            additional_options += ["-video_track_timescale", "1k"]

        ffmpeg.run_ffmpeg(
            [
                *ffmpeg_input_options,
                *additional_options,
                *ffmpeg_codecs_options,
                output_path,
            ]
        )


def _compose_concat_file(segment_paths, temp_directory, suffix: str = ""):
    if suffix:
        concat_file_path = Path(temp_directory, f"concat_{suffix}")
    else:
        concat_file_path = Path(temp_directory, "concat")

    concat_file_lines = [f"file '{path}'\n" for path in segment_paths]

    with open(concat_file_path, "w") as f:
        f.writelines(concat_file_lines)

    return concat_file_path


def ensure_cleanup_if_needed(f):
    @functools.wraps(f)
    def g(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        finally:
            if kwargs.get("cleanup", True):
                assert getattr(g, "paths_to_cleanup", None) is not None
                for path in g.paths_to_cleanup:
                    path.unlink(missing_ok=True)

    return g


@ensure_cleanup_if_needed
def merge_segments(
    audio_segment_paths: list[Path] | None = None,
    video_segment_paths: list[Path] | None = None,
    output_directory: str | Path | None = None,
    output_stem: str | Path | None = None,
    temp_directory: str | Path | None = None,
    cut_at_start: float = 0,
    cut_at_end: float = 0,
    metadata_tags: dict[str, str] = None,
    cleanup: bool = True,
) -> Path:
    """Merges and cuts media segments.

    The ``cut_at_start`` and ``cut_at_end`` arguments define the values to be
    passed to FFmpeg: ``-ss <cut_at_start>s`` and ``-to <cut_at_end>s``,
    respectively.

    Args:
        audio_segment_paths: Paths to audio segments.
        video_segment_paths: Paths to video segments.
        output_directory: Where to output the merged file.
        output_stem: An output stem of the merged file.
        temp_directory: Where to store intermediate files.
        cut_at_start: An offset (in seconds) to cut at start.
        cut_at_end: An offset (in seconds) to cut at end.
        metadata_tags: Metadata tags to write to the merged file.
        cleanup: Whether to cleanup intermediate files.

    Returns:
        A path of the merged file.
    """
    merge_segments.paths_to_cleanup = []

    if audio_segment_paths is None and video_segment_paths is None:
        raise ValueError("Audio or/and video paths should be provided")

    audio_segment_paths = audio_segment_paths or []
    video_segment_paths = video_segment_paths or []

    if audio_segment_paths and video_segment_paths:
        if len(audio_segment_paths) != len(video_segment_paths):
            raise ValueError("Audio and video paths are not of equal size")
        output_extension = ".mkv"
    elif audio_segment_paths:
        output_extension = os.path.splitext(audio_segment_paths[0])[1]
    elif video_segment_paths:
        output_extension = os.path.splitext(video_segment_paths[0])[1]

    if output_directory:
        output_directory = Path(output_directory)
    else:
        output_directory = Path.cwd()
    output_path = output_directory / f"{output_stem}{output_extension}"

    metadata_options: list[str] = []
    if metadata_tags:
        metadata_options = ["-movflags", "use_metadata_tags", "-map_metadata", "0"]
        for k, v in metadata_tags.items():
            metadata_options.extend(["-metadata", f"{k}={v}"])

    safe_concat_options = ["-safe", "0", "-f", "concat"]

    should_cut = bool(cut_at_start or cut_at_end)
    if should_cut is False:
        no_cut_concat_options = []
        if video_segment_paths:
            concat_file_path = _compose_concat_file(
                video_segment_paths, temp_directory, suffix="video"
            )
            merge_segments.paths_to_cleanup.append(concat_file_path)
            no_cut_concat_options += safe_concat_options + ["-i", concat_file_path]
        if audio_segment_paths:
            concat_file_path = _compose_concat_file(
                audio_segment_paths, temp_directory, suffix="audio"
            )
            merge_segments.paths_to_cleanup.append(concat_file_path)
            no_cut_concat_options += safe_concat_options + ["-i", concat_file_path]
        ffmpeg.run_ffmpeg(
            [*no_cut_concat_options, "-c", "copy", *metadata_options, output_path]
        )
    else:
        num_of_segments = len(audio_segment_paths or video_segment_paths)

        start_audio_segment_path = get_nth_or_none(audio_segment_paths, 0)
        start_video_segment_path = get_nth_or_none(video_segment_paths, 0)

        end_audio_segment_path = get_nth_or_none(audio_segment_paths, -1)
        end_video_segment_path = get_nth_or_none(video_segment_paths, -1)

        video_codec: str | None = None
        if video_segment_paths:
            with av.open(video_segment_paths[0]) as container:
                video_av_stream = container.streams.get({"video": 0})[0]
                video_codec = video_av_stream.codec_context.name

        if num_of_segments == 1:
            segment_trimmed_path = Path(temp_directory, "a.a" + output_extension)
            mux_and_cut_boundary_segment(
                start_audio_segment_path,
                start_video_segment_path,
                segment_trimmed_path,
                video_codec,
                cut_at_start=cut_at_start,
            )
            parts_to_merge = [segment_trimmed_path]
        elif num_of_segments == 2:
            start_trimmed_path = Path(temp_directory, "ab.a" + output_extension)
            mux_and_cut_boundary_segment(
                start_audio_segment_path,
                start_video_segment_path,
                start_trimmed_path,
                video_codec,
                cut_at_start=cut_at_start,
            )

            end_trimmed_path = start_trimmed_path.with_stem("ab.b")
            mux_and_cut_boundary_segment(
                end_audio_segment_path,
                end_video_segment_path,
                end_trimmed_path,
                video_codec,
                cut_at_end=cut_at_end,
            )

            parts_to_merge = [start_trimmed_path, end_trimmed_path]
        else:
            start_trimmed_path = Path(temp_directory, "abc.a" + output_extension)
            mux_and_cut_boundary_segment(
                start_audio_segment_path,
                start_video_segment_path,
                start_trimmed_path,
                video_codec,
                cut_at_start=cut_at_start,
            )

            middle_concatenated_path = start_trimmed_path.with_stem("abc.b")

            end_trimmed_path = start_trimmed_path.with_stem("abc.c")
            mux_and_cut_boundary_segment(
                end_audio_segment_path,
                end_video_segment_path,
                end_trimmed_path,
                video_codec,
                cut_at_end=cut_at_end,
            )

            bounds_slice = slice(1, -1)
            concat_filter_options = []

            if video_segment_paths:
                concat_file_path = _compose_concat_file(
                    video_segment_paths[bounds_slice],
                    temp_directory,
                    suffix="video",
                )
                concat_filter_options += safe_concat_options + ["-i", concat_file_path]
                merge_segments.paths_to_cleanup.append(concat_file_path)

            if audio_segment_paths:
                concat_file_path = _compose_concat_file(
                    audio_segment_paths[bounds_slice],
                    temp_directory,
                    suffix="audio",
                )
                concat_filter_options += safe_concat_options + ["-i", concat_file_path]
                merge_segments.paths_to_cleanup.append(concat_file_path)

            additional_options: list[str] = []
            if video_segment_paths and video_codec == "h264":
                additional_options += ["-video_track_timescale", "1k"]

            ffmpeg.run_ffmpeg(
                [
                    *concat_filter_options,
                    *additional_options,
                    "-c",
                    "copy",
                    middle_concatenated_path,
                ],
                capture_output=True,
                check=True,
            )

            parts_to_merge = [
                start_trimmed_path,
                middle_concatenated_path,
                end_trimmed_path,
            ]

        merge_segments.paths_to_cleanup.extend(parts_to_merge)

        if num_of_segments == 1:
            ffmpeg.run_ffmpeg(
                [
                    "-i",
                    parts_to_merge[0],
                    "-c",
                    "copy",
                    *metadata_options,
                    output_path,
                ]
            )
        else:
            concat_file_path = _compose_concat_file(parts_to_merge, temp_directory)
            ffmpeg.run_ffmpeg(
                [
                    *safe_concat_options,
                    "-i",
                    concat_file_path,
                    "-c",
                    "copy",
                    *metadata_options,
                    output_path,
                ]
            )
            merge_segments.paths_to_cleanup.append(concat_file_path)

    return output_path



"""Download media segments."""

import io
import os
from pathlib import Path
from typing import Callable, Generator
from urllib.parse import urljoin

import requests
import structlog

from ytpb.errors import SegmentDownloadError
from ytpb.types import SegmentSequence
from ytpb.utils.url import extract_media_type_from_url, extract_parameter_from_url

logger = structlog.get_logger(__name__)

SegmentOutputFilename = str | Callable[[SegmentSequence, str], str]


def _request_segment(
    sequence: SegmentSequence,
    base_url: str,
    size: int | None = None,
    session: requests.Session | None = None,
) -> requests.Response:
    get_function = session.get if session else requests.get

    headers = {}
    if size:
        headers["Range"] = f"bytes=0-{size}"

    response = get_function(urljoin(base_url, f"sq/{sequence}"), headers=headers)

    try:
        response.raise_for_status()
    except requests.exceptions.HTTPError:
        exception = SegmentDownloadError(
            f"Failed to download segment {sequence}: {response.reason}", sequence
        )
        logger.error(exception, exc_info=True)
        raise exception from None

    if size and response.status_code == requests.codes.ok:
        logger.debug("Header 'Range' is ignored, downloaded all content")

    return response


def compose_default_segment_filename(sequence: SegmentSequence, base_url: str) -> str:
    itag = extract_parameter_from_url("itag", base_url)
    extension = extract_media_type_from_url(base_url)[1]
    return f"{sequence}.i{itag}.{extension}"


def save_segment_to_file(
    content: bytes,
    sequence: SegmentSequence,
    base_url: str,
    output_directory: Path,
    output_filename,
) -> Path:
    if callable(output_filename):
        output_filename_value = output_filename(sequence, base_url)
        output_path = Path(output_directory) / output_filename_value
    else:
        output_path = Path(output_directory) / output_filename

    with open(output_path, "wb") as f:
        f.write(content)

    return output_path


def download_segment(
    sequence: int,
    base_url: str,
    output_directory: Path,
    output_filename: SegmentOutputFilename = compose_default_segment_filename,
    size: int | None = None,
    session: requests.Session | None = None,
    force_download: bool = True,
) -> Path:
    """Downloads a segment to file.

    Args:
        sequence: A segment sequence number.
        base_url: A segment base URL.
        output_directory: Where to download a segment.
        output_filename: A segment output filename.
        size: An amount of bytes to download.
        session: A :class:`requests.Session` object.
        force_download: Whether to download a segment if it already exists.

    Returns:
        A path where a segment was downloaded.

    Raises:
        SegmentDownloadError: If failed to download a segment.
    """
    if callable(output_filename):
        output_filename_value = output_filename(sequence, base_url)
        path_to_download_to = Path(output_directory) / output_filename_value
    else:
        path_to_download_to = Path(output_directory) / output_filename

    if force_download or not os.path.isfile(path_to_download_to):
        with open(path_to_download_to, "wb") as f:
            response = _request_segment(sequence, base_url, size, session)
            f.write(response.content)

    return path_to_download_to


def download_segment_to_buffer(
    sequence: SegmentSequence,
    base_url: str,
    size: int | None = None,
    session: requests.Session | None = None,
) -> io.BytesIO:
    """Downloads a segment to buffer.

    Args:
        sequence: A segment sequence number.
        base_url: A segment base URL.
        size: An amount of bytes to download.
        session: A :class:`requests.Session` object.

    Returns:
        An :class:`io.BytesIO` object.

    Raises:
        SegmentDownloadError: If failed to download a segment.
    """
    response = _request_segment(sequence, base_url, size, session)
    return io.BytesIO(response.content)


def iter_segments(
    sequences: list[SegmentSequence],
    base_url: str,
    size: int | None = None,
    session: requests.Session | None = None,
) -> Generator[tuple[requests.Response, SegmentSequence, str], None, None]:
    """Iterates over segment sequence numbers and requests segments.

    Args:
        sequences: Segment sequence numbers.
        base_url: A segment base URL.
        size: An amount of bytes to download.
        session: A :class:`requests.Session` object.

    Yields:
        Tuples of a :class:`requests.Response` object, segment sequence number,
        and base URL.
    """
    for sequence in sequences:
        with _request_segment(sequence, base_url, size, session) as response:
            yield response, sequence, base_url



import email
import os
import sys
from datetime import datetime, timedelta
from pathlib import Path

import click
import cloup
import structlog

from ytpb.cli.utils.date import format_timedelta, round_date

from ytpb.errors import (
    BadCommandArgument,
    BroadcastStatusError,
    CachedItemNotFoundError,
    MaxRetryError,
    QueryError,
    SegmentDownloadError,
)
from ytpb.fetchers import YoutubeDLInfoFetcher, YtpbInfoFetcher

from ytpb.format_spec import query_items
from ytpb.info import BroadcastStatus
from ytpb.playback import Playback, RewindInterval
from ytpb.types import DateInterval, SegmentSequence, SetOfStreams
from ytpb.utils.url import extract_parameter_from_url, normalize_video_url

logger = structlog.getLogger(__name__)

CONSOLE_TEXT_WIDTH = 80
EARLIEST_DATE_TIMEDELTA = timedelta(days=6, hours=23)


def normalize_stream_url(ctx: click.Context, param: click.Argument, value: str) -> str:
    try:
        return normalize_video_url(value)
    except BadCommandArgument as e:
        raise click.BadParameter(str(e))


stream_argument = cloup.argument(
    "stream_url",
    metavar="STREAM",
    help="Stream URL or video ID.",
    callback=normalize_stream_url,
)


def suppress_output() -> None:
    sys.stdout = open(os.devnull, "w")


def echo_notice(message: str) -> None:
    click.echo(f"~ {message}")


def get_parameter_by_name(name: str, ctx: click.Context) -> click.Parameter:
    return next((p for p in ctx.command.params if p.name == name), None)


def resolve_output_path(output_path: Path) -> Path:
    resolved_path = Path(output_path).expanduser().resolve()
    resolved_path.parent.mkdir(parents=True, exist_ok=True)
    return resolved_path


def prepare_line_for_summary_info(
    date: datetime,
    input_actual_timedelta: timedelta,
    use_ms_precision: bool = False,
) -> str:
    if not use_ms_precision:
        date = round_date(date)
    date_diff_string = format_timedelta(input_actual_timedelta, use_ms_precision)

    formatted = email.utils.format_datetime(date).split(", ")[1]
    if use_ms_precision:
        date_part, tz_part = formatted[:-6], formatted[-5:]
        ms_value = f"{date.microsecond // 1000:03g}"
        formatted = f"{date_part}.{ms_value} {tz_part}"

    if date_diff_string[1:] != "00:00":
        output = f"{formatted} ({date_diff_string})"
    else:
        output = formatted

    return output


def print_summary_info(
    input_date_interval: DateInterval,
    actual_date_interval: DateInterval,
    rewind_interval: RewindInterval,
    use_ms_precision: bool = False,
) -> None:
    input_tzinfo = input_date_interval.start.tzinfo

    start_time_info_line = prepare_line_for_summary_info(
        actual_date_interval.start.astimezone(input_tzinfo),
        actual_date_interval.start - input_date_interval.start,
        use_ms_precision,
    )
    click.echo(
        "Actual start: {}, seq. {}".format(
            start_time_info_line, rewind_interval.start.sequence
        )
    )

    end_time_info_line = prepare_line_for_summary_info(
        actual_date_interval.end.astimezone(input_tzinfo),
        actual_date_interval.end - input_date_interval.end,
        use_ms_precision,
    )
    click.echo(
        "  Actual end: {}, seq. {}".format(
            end_time_info_line, rewind_interval.end.sequence
        )
    )

    if rewind_interval.start.falls_in_gap:
        echo_notice("Start point is shifted right due to a stream gap!")
    if rewind_interval.end.falls_in_gap:
        echo_notice("End point is shifted left due to a stream gap!")


def create_playback(ctx: click.Context) -> Playback:
    stream_url = ctx.params["stream_url"]

    if ctx.params["yt_dlp"]:
        fetcher = YoutubeDLInfoFetcher(stream_url)
    else:
        fetcher = YtpbInfoFetcher(stream_url)

    try:
        click.echo(f"Run playback for {stream_url}")
        click.echo("(<<) Collecting info about the video...")

        force_update_cache = ctx.params.get("force_update_cache", False)
        need_read_cache = not (ctx.params.get("no_cache", True) or force_update_cache)
        if need_read_cache:
            try:
                playback = Playback.from_cache(stream_url, fetcher=fetcher)
            except CachedItemNotFoundError:
                logger.debug("Could not find unexpired cached item for the video")
                playback = Playback.from_url(
                    stream_url, fetcher=fetcher, write_to_cache=True
                )
        else:
            playback = Playback.from_url(
                stream_url, fetcher=fetcher, write_to_cache=force_update_cache
            )
    except BroadcastStatusError as e:
        match e.status:
            case BroadcastStatus.NONE:
                click.echo("It's seems that the video is not a live stream", err=True)
            case BroadcastStatus.COMPLETED:
                click.echo("Stream was live, but now it's finished", err=True)
        sys.exit(1)

    click.echo(f"Stream '{playback.info.title}' is alive!")

    return playback


def query_streams_or_exit(
    streams: SetOfStreams,
    format_spec: str,
    param: str | None = None,
    allow_empty: bool = False,
    allow_many: bool = True,
) -> SetOfStreams:
    try:
        queried_streams = query_items(format_spec, streams)
    except QueryError as exc:
        message = f"error: Invalid value for '{param}'. {exc}: '{format_spec}'"
        click.echo(message, err=True)
        sys.exit(1)

    if queried_streams:
        if len(queried_streams) > 1 and not allow_many:
            logger.error(
                "Too many queried streams", itags=[s.itag for s in queried_streams]
            )
            click.echo(
                (
                    f"error: Format spec is ambiguous: '{format_spec}'.\n\n"
                    "Found more than one stream matched a format spec. Please be more "
                    "explicit, or try 'yt-dlp --live-from-start -F ...' first."
                )
            )
            sys.exit(1)
    elif not allow_empty:
        message = (
            f"error: No streams found matching {param} format spec: '{format_spec}'"
        )
        click.echo(message, err=True)
        sys.exit(1)

    return queried_streams


def raise_for_too_far_sequence(
    sequence: int,
    current_sequence: int,
    base_url: str,
    ctx: click.Context,
    param_name: str,
) -> None:
    segment_duration = int(float(extract_parameter_from_url("dur", base_url)))
    earliest_sequence = (
        current_sequence - EARLIEST_DATE_TIMEDELTA.total_seconds() // segment_duration
    )
    if sequence <= earliest_sequence:
        click.echo("\n")
        raise click.BadParameter(
            f"Sequence number is beyond the limit of 7 days: {sequence}",
            param=get_parameter_by_name(param_name, ctx),
        )


def raise_for_sequence_ahead_of_current(
    sequence, current_sequence: int, ctx: click.Context, param_name: str
) -> None:
    if sequence >= current_sequence:
        click.echo("\n")
        message = (
            "Sequence number is ahead or equal to the current one: "
            f"{sequence} >= {current_sequence}"
        )
        param = get_parameter_by_name(param_name, ctx)
        raise click.BadParameter(message, ctx, param)


def find_earliest_sequence(
    playback: Playback, head_sequence: SegmentSequence, head_date: datetime
) -> SegmentSequence:
    """Finds the earliest available segment sequence number.

    Note that the function returns not the most earliest available sequence
    number, but the safe one, which is close to the 7-days availability limit
    (or potentially to the lower limit if a stream is stored less, for example,
    5 days, if such a thing occurs).
    """
    some_stream = next(iter(playback.streams))
    segment_duration = float(extract_parameter_from_url("dur", some_stream.base_url))

    target_date = head_date - EARLIEST_DATE_TIMEDELTA
    logger.info("Start finding the earliest segment", limit=target_date)

    jump_to_left = int(EARLIEST_DATE_TIMEDELTA.total_seconds() // segment_duration)
    left_sequence = head_sequence - jump_to_left
    right_sequence = head_sequence - jump_to_left // 2
    if left_sequence < 0:
        logger.debug("Found beginning segment to be the earliest")
        return 0

    search_domain = range(left_sequence, right_sequence + 1)
    logger.debug("Set search domain", left=left_sequence, right=right_sequence)

    def _check_ahead_target(sequence: SegmentSequence, target_date: datetime) -> bool:
        try:
            segment = playback.get_segment(sequence, some_stream)
        except (SegmentDownloadError, MaxRetryError):
            timedelta_to_target = None
            result = False
        else:
            timedelta_to_target = segment.ingestion_start_date - target_date
            if timedelta_to_target > timedelta(0):
                result = True
            else:
                result = False
        logger.debug("Bisect step", sequence=sequence, timedelta=timedelta_to_target)
        return result

    low, high = 0, len(search_domain)
    while low < high:
        middle = (low + high) // 2
        candidate_sequence = search_domain[middle]
        if _check_ahead_target(candidate_sequence, target_date):
            high = middle
        else:
            low = middle + 1

    result = search_domain[low]
    logger.debug("Found the earliest segment", sequence=result)

    return result
